# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

from typing import Tuple
import numpy as np


from ft.onto.base_ontology import Annotation
from forte.utils import create_import_error_msg
from forte.common.configuration import Config
from forte.processors.data_augment.algorithms.text_replacement_op import (
    TextReplacementOp,
)

__all__ = [
    "EmbeddingSimilarityReplacementOp",
]


class EmbeddingSimilarityReplacementOp(TextReplacementOp):
    r"""
    This class is a replacement op leveraging pre-trained word
    embeddings, such as `word2vec` and `glove`, to replace the input
    word with another word with similar word embedding.
    By default, the replacement word is randomly chosen from the
    top k words with the most similar embeddings.

    Args:
        configs:
            The config should contain the following key-value pairs:

            - `vocab_path` (str):
                The absolute path to the vocabulary file for
                the pretrained embeddings

            - `embed_hparams` (dict):
                The hyper-parameters to initialize the
                texar.torch.data.Embedding object.

            - `top_k` (int):
                the number of k most similar words to choose from
    """

    def __init__(self, configs: Config):
        super().__init__(configs)
        try:
            from texar.torch.data import (  # pylint:disable=import-outside-toplevel
                Vocab,
                Embedding,
            )
        except ImportError as e:
            raise ImportError(
                create_import_error_msg(
                    "texar-pytorch",
                    "data_aug",
                    "EmbeddingSimilarityReplacementOp",
                )
            ) from e

        self.vocab = Vocab(self.configs["vocab_path"])
        embed_hparams = self.configs["embed_hparams"]
        embedding = Embedding(self.vocab.token_to_id_map_py, embed_hparams)
        self.normalized_vectors = (
            embedding.word_vecs
            / np.sqrt((embedding.word_vecs**2).sum(axis=1))[:, np.newaxis]
        )

    def replace(self, input_anno: Annotation) -> Tuple[bool, str]:
        r"""
        This function replaces a word words with similar
        pretrained embeddings.

        Args:
            input_anno: The input annotation.
        Returns:
            A tuple of two values, where the first element is a boolean value
            indicating whether the replacement happens, and the second
            element is the replaced word.
        """
        word = input_anno.text
        if word not in self.vocab.token_to_id_map_py:
            return False, word

        source_id = self.vocab.token_to_id_map_py[word]
        source_vector = self.normalized_vectors[source_id]
        scores = np.dot(self.normalized_vectors, source_vector)
        target_ids = np.argpartition(-scores, self.configs["top_k"] + 1)[
            : self.configs["top_k"] + 1
        ]
        target_words = [
            self.vocab.id_to_token_map_py[idx]
            for idx in target_ids
            if idx != source_id
            and self.vocab.id_to_token_map_py[idx].lower() != word.lower()
        ]
        return True, random.choice(target_words)



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Data augmentation operation for the Random Word Splitting operation.
Randomly choose n words (With length greater than 1) and split it at a random
position. Do this n times, where n = alpha * input length.
Example: Original Text -> "I will be there soon." ,
Augmented Text -> "I w ill be there so on."
"""

from math import ceil
import random
from typing import List, Iterable

from forte.data.data_pack import DataPack
from forte.data.ontology import Annotation
from forte.processors.data_augment.algorithms.base_data_augmentation_op import (
    BaseDataAugmentationOp,
)
from forte.utils.utils import get_class

__all__ = ["RandomWordSplitDataAugmentOp"]


class RandomWordSplitDataAugmentOp(BaseDataAugmentationOp):
    r"""
    This class creates an operation to perform Random Word Splitting.
    It randomly chooses n words in a sentence and splits each word at
    a random position where n = alpha * input length.
    alpha indicates the percent of the words in a sentence that are changed.
    """

    def augment(self, data_pack: DataPack) -> bool:
        r"""
        This function splits a given word at a random position and replaces
        the original word with 2 split parts of it.
        """
        augment_entry = get_class(self.configs["augment_entry"])

        annotations: List[Annotation] = []
        indexes: List[int] = []
        endings = []
        annos: Iterable[Annotation] = data_pack.get(augment_entry)
        try:
            for idx, anno in enumerate(annos):
                annotations.append(anno)
                indexes.append(idx)
                endings.append(anno.end)
            if len(annotations) > 0:
                annotation_to_split = random.sample(
                    [
                        (anno, idx)
                        for (anno, idx) in zip(annotations, indexes)
                        if (anno.end - anno.begin) > 1
                    ],
                    ceil(self.configs["alpha"] * len(annotations)),
                )
                annotation_to_split = sorted(
                    annotation_to_split, key=lambda x: x[1], reverse=True
                )
                for curr_anno in annotation_to_split:
                    src_anno, src_idx = curr_anno
                    splitting_position = random.randrange(
                        1, (src_anno.end - src_anno.begin)
                    )
                    word_split = [
                        src_anno.text[:splitting_position],
                        src_anno.text[splitting_position:],
                    ]
                    if src_idx != 0:
                        first_position = endings[src_idx - 1] + 1
                        second_position = endings[src_idx]
                        word_split[1] = " " + word_split[1]
                    else:
                        first_position = 0
                        second_position = endings[0]
                        word_split[1] = " " + word_split[1]

                    self.insert_annotated_span(
                        word_split[1],
                        data_pack,
                        second_position,
                        self.configs["augment_entry"],
                    )

                    self.delete_annotation(src_anno)

                    self.insert_annotated_span(
                        word_split[0],
                        data_pack,
                        first_position,
                        self.configs["augment_entry"],
                    )
            return True
        except ValueError:
            return False

    @classmethod
    def default_configs(cls):
        """
        Returns:
            A dictionary with the default config for this processor.

        Additional keys for determining how many words will be split:
            - alpha (float):
                0 <= alpha <= 1. indicates the percent of the words
                in a sentence that are changed.
            - augment_entry (str):
                Defines the entry the processor will augment.
                It should be a full qualified name of the entry class.
                For example, "ft.onto.base_ontology.Sentence".
        """
        return {
            "augment_entry": "ft.onto.base_ontology.Token",
            "alpha": 0.1,
        }



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
In this program we try to implement the famous Eliza chatbot, a rule-based
chatbot invented in 1964. The rule-based and model-less nature makes it
extremely suitable for demonstration purposes. Note that this implementation
may not reproduce the exact behavior of the original Eliza, instead, we are
largely based on the implementation here: https://github.com/wadetb/eliza

For more information of Eliza, see https://en.wikipedia.org/wiki/ELIZA and
the paper at: https://dl.acm.org/doi/10.1145/365153.365168
"""
import logging
import random
import re
from typing import List, Dict, Union, Optional

from forte.data.common_entry_utils import get_last_utterance, create_utterance
from forte.data.data_pack import DataPack
from forte.processors.base import PackProcessor
from ft.onto.base_ontology import Utterance

__all__ = [
    "ElizaProcessor",
]


class Key:
    def __init__(self, word, weight, decomps):
        self.word = word
        self.weight = weight
        self.decomps = decomps


class Decomp:
    def __init__(self, parts, save, reasmbs):
        self.parts = parts
        self.save = save
        self.reasmbs = reasmbs
        self.next_reasmb_index = 0


class ElizaProcessor(PackProcessor):
    """
    The ElizaProcessor adapted from https://github.com/wadetb/eliza. This
    processor will generate response based on the Eliza rules. For more
    information, please refer to https://dl.acm.org/doi/10.1145/365153.365168.

    This processor is not parallel ready because it will keep an internal
    state `memory`.

    """

    def __init__(self):
        super().__init__()
        self.initials: List[str] = []
        self.finals: List[str] = []
        self.quits: List[str] = []
        self.pres: Dict[str, List[str]] = {}
        self.posts: Dict[str, List[str]] = {}
        self.synons: Dict[str, List[str]] = {}
        self.keys: Dict[str, Key] = {}
        self.memory: List[str] = []

        self.parse_doctor()

    def parse_doctor(self):
        parse_nodes: List[Union[Key, Decomp]] = []

        for line in __doctor__.splitlines():
            if not line.strip():
                continue

            depth = int((len(line) - len(line.lstrip())) / 2)
            parse_nodes = parse_nodes[:depth]

            tag, content = [part.strip() for part in line.split(":")]

            if len(parse_nodes) == 0:
                if tag == "initial":
                    self.initials.append(content)
                elif tag == "final":
                    self.finals.append(content)
                elif tag == "quit":
                    self.quits.append(content)
                elif tag == "pre":
                    parts = content.split(" ")
                    self.pres[parts[0]] = parts[1:]
                elif tag == "post":
                    parts = content.split(" ")
                    self.posts[parts[0]] = parts[1:]
                elif tag == "synon":
                    parts = content.split(" ")
                    self.synons[parts[0]] = parts
                elif tag == "key":
                    parts = content.split(" ")
                    word = parts[0]
                    weight = int(parts[1]) if len(parts) > 1 else 1
                    key = Key(word, weight, [])
                    self.keys[word] = key
                    parse_nodes.append(key)
            else:
                parent_node = parse_nodes[-1]
                if isinstance(parent_node, Key):
                    if tag == "decomp":
                        parts = content.split(" ")
                        save = False
                        if parts[0] == "$":
                            save = True
                            parts = parts[1:]
                        decomp = Decomp(parts, save, [])
                        parent_node.decomps.append(decomp)
                        parse_nodes.append(decomp)
                    else:
                        raise ValueError(f"Unexpected parent node {tag}")
                if isinstance(parent_node, Decomp):
                    if tag == "reasmb":
                        parts = content.split(" ")
                        parent_node.reasmbs.append(parts)
                    else:
                        raise ValueError(f"Unexpected parent node {tag}")

    def _match_decomp_r(self, parts, words, results):
        if not parts and not words:
            return True
        if not parts or (not words and parts != ["*"]):
            return False
        if parts[0] == "*":
            for index in range(len(words), -1, -1):
                results.append(words[:index])
                if self._match_decomp_r(parts[1:], words[index:], results):
                    return True
                results.pop()
            return False
        elif parts[0].startswith("@"):
            root = parts[0][1:]
            if root not in self.synons:
                raise ValueError("Unknown synonym root {}".format(root))
            if not words[0].lower() in self.synons[root]:
                return False
            results.append([words[0]])
            return self._match_decomp_r(parts[1:], words[1:], results)
        elif parts[0].lower() != words[0].lower():
            return False
        else:
            return self._match_decomp_r(parts[1:], words[1:], results)

    def _match_decomp(self, parts, words):
        results = []
        if self._match_decomp_r(parts, words, results):
            return results
        return None

    def _next_reasmb(self, decomp):
        index = decomp.next_reasmb_index
        result = decomp.reasmbs[index % len(decomp.reasmbs)]
        decomp.next_reasmb_index = index + 1
        return result

    def _reassemble(self, reasmb, results):
        output = []
        for reword in reasmb:
            if not reword:
                continue
            if reword[0] == "(" and reword[-1] == ")":
                index = int(reword[1:-1])
                if index < 1 or index > len(results):
                    raise ValueError("Invalid result index {}".format(index))
                insert = results[index - 1]
                for punct in [",", ".", ";"]:
                    if punct in insert:
                        insert = insert[: insert.index(punct)]
                output.extend(insert)
            else:
                output.append(reword)
        return output

    def _sub(self, words, sub):
        output = []
        for word in words:
            word_lower = word.lower()
            if word_lower in sub:
                output.extend(sub[word_lower])
            else:
                output.append(word)
        return output

    def _match_key(self, words, key):
        for decomp in key.decomps:
            results = self._match_decomp(decomp.parts, words)
            if results is None:
                logging.debug("Decomp did not match: %s", decomp.parts)
                continue
            logging.debug("Decomp matched: %s", decomp.parts)
            logging.debug("Decomp results: %s", results)
            results = [self._sub(words, self.posts) for words in results]
            logging.debug("Decomp results after posts: %s", results)
            reasmb = self._next_reasmb(decomp)
            logging.debug("Using reassembly: %s", reasmb)
            if reasmb[0] == "goto":
                goto_key = reasmb[1]
                if goto_key not in self.keys:
                    raise ValueError("Invalid goto key {}".format(goto_key))
                logging.debug("Goto key: %s", goto_key)
                return self._match_key(words, self.keys[goto_key])
            output = self._reassemble(reasmb, results)
            if decomp.save:
                self.memory.append(output)
                logging.debug("Saved to memory: %s", output)
                continue
            return output
        return None

    def _respond(self, text: str):
        """
        Create a response based on the strategy.

        Args:
            text:

        Returns:

        """
        if text.lower() in self.quits:
            return None

        text = re.sub(r"\s*\.+\s*", " . ", text)
        text = re.sub(r"\s*,+\s*", " , ", text)
        text = re.sub(r"\s*;+\s*", " ; ", text)
        logging.debug("After punctuation cleanup: %s", text)

        words = [w for w in text.split(" ") if w]
        logging.debug("Input: %s", words)

        words = self._sub(words, self.pres)
        logging.debug("After pre-substitution: %s", words)

        keys = [self.keys[w.lower()] for w in words if w.lower() in self.keys]
        keys = sorted(keys, key=lambda k: -k.weight)
        logging.debug("Sorted keys: %s", [(k.word, k.weight) for k in keys])

        output = None

        for key in keys:
            output = self._match_key(words, key)
            if output:
                logging.debug("Output from key: %s", output)
                break
        if not output:
            if self.memory:
                index = random.randrange(len(self.memory))
                output = self.memory.pop(index)
                logging.debug("Output from memory: %s", output)
            else:
                output = self._next_reasmb(self.keys["xnone"].decomps[0])
                logging.debug("Output from xnone: %s", output)

        return " ".join(output)

    def _process(self, input_pack: DataPack):
        utterance: Optional[Utterance] = get_last_utterance(input_pack, "user")
        if utterance is None:
            logging.info("Cannot get new user utterance. Showing initials.")
            create_utterance(input_pack, random.choice(self.initials), "ai")
        else:
            output = self._respond(utterance.text)

            if output is None:
                create_utterance(input_pack, random.choice(self.finals), "ai")
            else:
                create_utterance(input_pack, output, "ai")


# The doctor for initializing Eliza. Also from https://github.com/wadetb/eliza.
__doctor__ = """
initial: How do you do.  Please tell me your problem.
final: Goodbye.  Thank you for talking to me.
quit: bye
quit: goodbye
quit: quit
pre: dont don't
pre: cant can't
pre: wont won't
pre: recollect remember
pre: dreamt dreamed
pre: dreams dream
pre: maybe perhaps
pre: how what
pre: when what
pre: certainly yes
pre: machine computer
pre: computers computer
pre: were was
pre: you're you are
pre: i'm i am
pre: same alike
post: am are
post: your my
post: me you
post: myself yourself
post: yourself myself
post: i you
post: you I
post: my your
post: i'm you are
synon: belief feel think believe wish
synon: family mother mom father dad sister brother wife children child
synon: desire want need
synon: sad unhappy depressed sick
synon: happy elated glad better
synon: cannot can't
synon: everyone everybody nobody noone
synon: be am is are was
key: xnone
  decomp: *
    reasmb: I'm not sure I understand you fully.
    reasmb: Please go on.
    reasmb: What does that suggest to you ?
    reasmb: Do you feel strongly about discussing such things ?
key: sorry
  decomp: *
    reasmb: Please don't apologise.
    reasmb: Apologies are not necessary.
    reasmb: I've told you that apologies are not required.
key: apologise
  decomp: *
    reasmb: goto sorry
key: remember 5
  decomp: * i remember *
    reasmb: Do you often think of (2) ?
    reasmb: Does thinking of (2) bring anything else to mind ?
    reasmb: What else do you recollect ?
    reasmb: Why do you recollect (2) just now ?
    reasmb: What in the present situation reminds you of (2) ?
    reasmb: What is the connection between me and (2) ?
  decomp: * do you remember *
    reasmb: Did you think I would forget (2) ?
    reasmb: Why do you think I should recall (2) now ?
    reasmb: What about (2) ?
    reasmb: goto what
    reasmb: You mentioned (2) ?
key: if 3
  decomp: * if *
    reasmb: Do you think its likely that (2) ?
    reasmb: Do you wish that (2) ?
    reasmb: What do you know about (2) ?
    reasmb: Really, if (2) ?
key: dreamed 4
  decomp: * i dreamed *
    reasmb: Really, (2) ?
    reasmb: Have you ever fantasized (2) while you were awake ?
    reasmb: Have you ever dreamed (2) before ?
    reasmb: goto dream
key: dream 3
  decomp: *
    reasmb: What does that dream suggest to you ?
    reasmb: Do you dream often ?
    reasmb: What persons appear in your dreams ?
    reasmb: Do you believe that dreams have something to do with your problems ?
key: perhaps
  decomp: *
    reasmb: You don't seem quite certain.
    reasmb: Why the uncertain tone ?
    reasmb: Can't you be more positive ?
    reasmb: You aren't sure ?
    reasmb: Don't you know ?
key: name 15
  decomp: *
    reasmb: I am not interested in names.
    reasmb: I've told you before, I don't care about names -- please continue.
key: deutsch
  decomp: *
    reasmb: goto xforeign
    reasmb: I told you before, I don't understand German.
key: francais
  decomp: *
    reasmb: goto xforeign
    reasmb: I told you before, I don't understand French.
key: italiano
  decomp: *
    reasmb: goto xforeign
    reasmb: I told you before, I don't understand Italian.
key: espanol
  decomp: *
    reasmb: goto xforeign
    reasmb: I told you before, I don't understand Spanish.
key: xforeign
  decomp: *
    reasmb: I speak only English.
key: hello
  decomp: *
    reasmb: How do you do.  Please state your problem.
    reasmb: Hi.  What seems to be your problem ?
key: computer 50
  decomp: *
    reasmb: Do computers worry you ?
    reasmb: Why do you mention computers ?
    reasmb: What do you think machines have to do with your problem ?
    reasmb: Don't you think computers can help people ?
    reasmb: What about machines worrys you ?
    reasmb: What do you think about machines ?
key: am
  decomp: * am i *
    reasmb: Do you believe you are (2) ?
    reasmb: Would you want to be (2) ?
    reasmb: Do you wish I would tell you you are (2) ?
    reasmb: What would it mean if you were (2) ?
    reasmb: goto what
  decomp: *
    reasmb: Why do you say 'am' ?
    reasmb: I don't understand that.
key: are
  decomp: * are you *
    reasmb: Why are you interested in whether I am (2) or not ?
    reasmb: Would you prefer if I weren't (2) ?
    reasmb: Perhaps I am (2) in your fantasies.
    reasmb: Do you sometimes think I am (2) ?
    reasmb: goto what
  decomp: * are *
    reasmb: Did you think they might not be (2) ?
    reasmb: Would you like it if they were not (2) ?
    reasmb: What if they were not (2) ?
    reasmb: Possibly they are (2) .
key: your
  decomp: * your *
    reasmb: Why are you concerned over my (2) ?
    reasmb: What about your own (2) ?
    reasmb: Are you worried about someone else's (2) ?
    reasmb: Really, my (2) ?
key: was 2
  decomp: * was i *
    reasmb: What if you were (2) ?
    reasmb: Do you think you were (2) ?
    reasmb: Were you (2) ?
    reasmb: What would it mean if you were (2) ?
    reasmb: What does (2) suggest to you ?
    reasmb: goto what
  decomp: * i was *
    reasmb: Were you really ?
    reasmb: Why do you tell me you were (2) now ?
    reasmb: Perhaps I already know you were (2) .
  decomp: * was you *
    reasmb: Would you like to believe I was (2) ?
    reasmb: What suggests that I was (2) ?
    reasmb: What do you think ?
    reasmb: Perhaps I was (2) .
    reasmb: What if I had been (2) ?
key: i
  decomp: * i @desire *
    reasmb: What would it mean to you if you got (3) ?
    reasmb: Why do you want (3) ?
    reasmb: Suppose you got (3) soon ?
    reasmb: What if you never got (3) ?
    reasmb: What would getting (3) mean to you ?
    reasmb: What does wanting (3) have to do with this discussion ?
  decomp: * i am * @sad *
    reasmb: I am sorry to hear that you are (3) .
    reasmb: Do you think that coming here will help you not to be (3) ?
    reasmb: I'm sure it's not pleasant to be (3) .
    reasmb: Can you explain what made you (3) ?
  decomp: * i am * @happy *
    reasmb: How have I helped you to be (3) ?
    reasmb: Has your treatment made you (3) ?
    reasmb: What makes you (3) just now ?
    reasmb: Can you explan why you are suddenly (3) ?
  decomp: * i was *
    reasmb: goto was
  decomp: * i @belief * i *
    reasmb: Do you really think so ?
    reasmb: But you are not sure you (3) .
    reasmb: Do you really doubt you (3) ?
  decomp: * i * @belief * you *
    reasmb: goto you
  decomp: * i am *
    reasmb: Is it because you are (2) that you came to me ?
    reasmb: How long have you been (2) ?
    reasmb: Do you believe it is normal to be (2) ?
    reasmb: Do you enjoy being (2) ?
  decomp: * i @cannot *
    reasmb: How do you think that you can't (3) ?
    reasmb: Have you tried ?
    reasmb: Perhaps you could (3) now.
    reasmb: Do you really want to be able to (3) ?
  decomp: * i don't *
    reasmb: Don't you really (2) ?
    reasmb: Why don't you (2) ?
    reasmb: Do you wish to be able to (2) ?
    reasmb: Does that trouble you ?
  decomp: * do i feel *
    reasmb: Tell me more about such feelings.
    reasmb: Do you often feel (2) ?
    reasmb: Do you enjoy feeling (2) ?
    reasmb: Of what does feeling (2) remind you ?
  decomp: * i * you *
    reasmb: Perhaps in your fantasies we (2) each other.
    reasmb: Do you wish to (2) me ?
    reasmb: You seem to need to (2) me.
    reasmb: Do you (2) anyone else ?
  decomp: *
    reasmb: You say (1) ?
    reasmb: Can you elaborate on that ?
    reasmb: Do you say (1) for some special reason ?
    reasmb: That's quite interesting.
key: you
  decomp: * you remind me of *
    reasmb: goto alike
  decomp: * you are *
    reasmb: What makes you think I am (2) ?
    reasmb: Does it please you to believe I am (2) ?
    reasmb: Do you sometimes wish you were (2) ?
    reasmb: Perhaps you would like to be (2) .
  decomp: * you * me *
    reasmb: Why do you think I (2) you ?
    reasmb: You like to think I (2) you -- don't you ?
    reasmb: What makes you think I (2) you ?
    reasmb: Really, I (2) you ?
    reasmb: Do you wish to believe I (2) you ?
    reasmb: Suppose I did (2) you -- what would that mean ?
    reasmb: Does someone else believe I (2) you ?
  decomp: * you *
    reasmb: We were discussing you -- not me.
    reasmb: Oh, I (2) ?
    reasmb: You're not really talking about me -- are you ?
    reasmb: What are your feelings now ?
key: yes
  decomp: *
    reasmb: You seem to be quite positive.
    reasmb: You are sure.
    reasmb: I see.
    reasmb: I understand.
key: no
  decomp: *
    reasmb: Are you saying no just to be negative?
    reasmb: You are being a bit negative.
    reasmb: Why not ?
    reasmb: Why 'no' ?
key: my 2
  decomp: $ * my *
    reasmb: Lets discuss further why your (2) .
    reasmb: Earlier you said your (2) .
    reasmb: But your (2) .
    reasmb: Does that have anything to do with the fact that your (2) ?
  decomp: * my * @family *
    reasmb: Tell me more about your family.
    reasmb: Who else in your family (4) ?
    reasmb: Your (3) ?
    reasmb: What else comes to mind when you think of your (3) ?
  decomp: * my *
    reasmb: Your (2) ?
    reasmb: Why do you say your (2) ?
    reasmb: Does that suggest anything else which belongs to you ?
    reasmb: Is it important that your (2) ?
key: can
  decomp: * can you *
    reasmb: You believe I can (2) don't you ?
    reasmb: goto what
    reasmb: You want me to be able to (2) .
    reasmb: Perhaps you would like to be able to (2) yourself.
  decomp: * can i *
    reasmb: Whether or not you can (2) depends on you more than me.
    reasmb: Do you want to be able to (2) ?
    reasmb: Perhaps you don't want to (2) .
    reasmb: goto what
key: what
  decomp: *
    reasmb: Why do you ask ?
    reasmb: Does that question interest you ?
    reasmb: What is it you really wanted to know ?
    reasmb: Are such questions much on your mind ?
    reasmb: What answer would please you most ?
    reasmb: What do you think ?
    reasmb: What comes to mind when you ask that ?
    reasmb: Have you asked such questions before ?
    reasmb: Have you asked anyone else ?
key: because
  decomp: *
    reasmb: Is that the real reason ?
    reasmb: Don't any other reasons come to mind ?
    reasmb: Does that reason seem to explain anything else ?
    reasmb: What other reasons might there be ?
key: why
  decomp: * why don't you *
    reasmb: Do you believe I don't (2) ?
    reasmb: Perhaps I will (2) in good time.
    reasmb: Should you (2) yourself ?
    reasmb: You want me to (2) ?
    reasmb: goto what
  decomp: * why can't i *
    reasmb: Do you think you should be able to (2) ?
    reasmb: Do you want to be able to (2) ?
    reasmb: Do you believe this will help you to (2) ?
    reasmb: Have you any idea why you can't (2) ?
    reasmb: goto what
  decomp: *
    reasmb: goto what
key: everyone 2
  decomp: * @everyone *
    reasmb: Realy, (2) ?
    reasmb: Surely not (2) .
    reasmb: Can you think of anyone in particular ?
    reasmb: Who, for example?
    reasmb: Are you thinking of a very special person ?
    reasmb: Who, may I ask ?
    reasmb: Someone special perhaps ?
    reasmb: You have a particular person in mind, don't you ?
    reasmb: Who do you think you're talking about ?
key: everybody 2
  decomp: *
    reasmb: goto everyone
key: nobody 2
  decomp: *
    reasmb: goto everyone
key: noone 2
  decomp: *
    reasmb: goto everyone
key: always 1
  decomp: *
    reasmb: Can you think of a specific example ?
    reasmb: When ?
    reasmb: What incident are you thinking of ?
    reasmb: Really, always ?
key: alike 10
  decomp: *
    reasmb: In what way ?
    reasmb: What resemblence do you see ?
    reasmb: What does that similarity suggest to you ?
    reasmb: What other connections do you see ?
    reasmb: What do you suppose that resemblence means ?
    reasmb: What is the connection, do you suppose ?
    reasmb: Could here really be some connection ?
    reasmb: How ?
key: like 10
  decomp: * @be * like *
    reasmb: goto alike
"""



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
RemoteProcessor is used to interact with a remote Forte end-point.
The Forte service must be created by a pipeline with `RawDataDeserializeReader`
being set as its reader.
"""

import json
import logging
from typing import Dict, Set, Any, Optional

from forte.common import Resources, ProcessorConfigError
from forte.common.configuration import Config
from forte.data.data_pack import DataPack
from forte.processors.base import PackProcessor
from forte.utils import create_import_error_msg


logger = logging.getLogger(__name__)

__all__ = ["RemoteProcessor"]


class RemoteProcessor(PackProcessor):
    r"""
    RemoteProcessor wraps up the interactions with remote Forte end point.
    Each input DataPack from the upstream component will be serialized and
    packed into a POST request to be sent to a remote service, which should
    return a response that can be parsed into a DataPack to update the input.
    Example usage:

    .. code-block:: python

        # Assume that a Forte service is running on "localhost:8080".
        Pipeline() \
            .set_reader(plaintext_reader(), {"input_path":"some/path"}) \
            .add(RemoteProcessor(), {"url": "http://localhost:8008"})

    """

    def __init__(self):
        super().__init__()
        try:
            import requests  # pylint: disable=import-outside-toplevel
        except ImportError as e:
            raise ImportError(
                create_import_error_msg(
                    "requests", "remote", "Remote Processor"
                )
            ) from e
        self._requests: Any = requests
        self._records: Optional[Dict[str, Set[str]]] = None
        self._expectation: Optional[Dict[str, Set[str]]] = None

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        _validation: Config = self.configs.validation

        # Verify the service is running
        response = self._requests.get(self.configs.url)
        if response.status_code != 200 or response.json()["status"] != "OK":
            raise ProcessorConfigError(
                f"{response.status_code} {response.reason}: Please double "
                "check your endpoint URL configuration and make sure that the "
                f"remote service at {self.configs.url} is a valid pipeline "
                "service that is up and running."
            )
        service_name: str = response.json()["service_name"]
        input_format: str = response.json()["input_format"]

        if _validation.do_init_type_check:
            # Validate service name and input format
            if service_name != _validation.expected_name:
                raise ProcessorConfigError(
                    "Validation fail: The expected service name "
                    f"('{_validation.expected_name}') does not match the "
                    "actual name returned by remote service "
                    f"('{service_name}'). Please double check your endpoint "
                    f"URL {self.configs.url} or consider updating the configs "
                    "of RemoteProcessor so that 'validation.expected_name' "
                    f"equals to '{service_name}'."
                )
            if input_format != _validation.input_format:
                raise ProcessorConfigError(
                    "Validation fail: The expected input format "
                    f"('{_validation.input_format}') does not match the "
                    "actual input format returned by remote service "
                    f"('{input_format}'). Please double check your endpoint "
                    f"URL {self.configs.url} or consider updating the configs "
                    "of RemoteProcessor so that 'validation.input_format' "
                    f"equals to '{input_format}'."
                )

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of `RemoteProcessor`. The records
        are queried from the remote service. The types and attributes are
        populated from all the components in remote pipeline.

        Args:
            record_meta: the field in the datapack for type record that need to
                fill in for consistency checking.
        """
        if self._records is None:
            response = self._requests.get(f"{self.configs.url}/records")
            if response.status_code != 200 or response.json()["status"] != "OK":
                raise ProcessorConfigError(
                    f"{response.status_code} {response.reason}: "
                    "Fail to fetch records from remote service. Please make "
                    f"sure that the remote service at {self.configs.url} is "
                    "a valid pipeline service that is up and running."
                )
            self._records = response.json()["records"]
        record_meta.update(self._records)

    def expected_types_and_attributes(self):
        r"""Method to add expected types and attributes for the input of
        `RemoteProcessor`. This should be the `expected_types_and_attributes`
        of the first processor in remote pipeline.
        """
        if self._expectation is None:
            response = self._requests.get(f"{self.configs.url}/expectation")
            if response.status_code != 200 or response.json()["status"] != "OK":
                raise ProcessorConfigError(
                    f"{response.status_code} {response.reason}: "
                    "Fail to fetch expected types and attributes from remote "
                    "service. Please make sure that the remote service at "
                    f"{self.configs.url} is a valid pipeline service that is "
                    "up and running."
                )
            self._expectation = response.json()["expectation"]
        return self._expectation

    def _process(self, input_pack: DataPack):
        # Pack the input_pack and POST it to remote service
        response = self._requests.post(
            f"{self.configs.url}/process",
            json={"args": json.dumps([[input_pack.to_string()]])},
        )
        if response.status_code != 200 or response.json()["status"] != "OK":
            raise Exception(
                f"{response.status_code} {response.reason}: "
                "Invalid post request to process input pack. Please make "
                f"sure that the remote service at {self.configs.url} is "
                "a valid pipeline service that is up and running."
            )
        result = response.json()["result"]
        input_pack.update(DataPack.from_string(result))  # type: ignore

    def set_test_mode(self, app):
        """
        Configure the processor into test mode. This should only be called
        from a pytest program.

        Args:
            app: A fastapi app from a Forte pipeline.
        """
        try:
            # pylint: disable=import-outside-toplevel
            from fastapi.testclient import TestClient
        except ImportError as err:
            raise ImportError(
                create_import_error_msg("fastapi", "remote", "RemoteProcessor")
            ) from err
        self._requests = TestClient(app)

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """
        This defines a basic config structure for RemoteProcessor.
        Following are the keys for this dictionary:

            - ``url``: URL of the remote service end point.
              Default value is `"http://localhost:8008"`.
            - ``validation``: Information for validation.

                - ``do_init_type_check``: Validate the pipeline by checking
                  the info of the remote pipeline with the expected
                  attributes. Default to `False`.
                - ``input_format``: The expected input format of the remote
                  service. Default to `"string"`.
                - ``expected_name``: The expected pipeline name.
                  Default to `''`.

        Returns:
            dict: A dictionary with the default config for this processor.
        """
        return {
            "url": "http://localhost:8008",
            "validation": {
                "do_init_type_check": False,
                "input_format": "string",
                "expected_name": "",
            },
        }



# ***automatically_generated***
# ***source json:forte/ontology_specs/wikipedia.json***
# flake8: noqa
# mypy: ignore-errors
# pylint: skip-file
"""
Automatically generated ontology wikipedia. Do not change manually.
"""

from dataclasses import dataclass
from forte.data.data_pack import DataPack
from forte.data.ontology.top import Annotation
from forte.data.ontology.top import Generics
from typing import List
from typing import Optional

__all__ = [
    "WikiPage",
    "WikiBody",
    "WikiSection",
    "WikiParagraph",
    "WikiTitle",
    "WikiArticleTitle",
    "WikiAnchor",
    "WikiInfoBoxProperty",
    "WikiInfoBoxMapped",
    "WikiCategory",
]


@dataclass
class WikiPage(Annotation):
    """
    Attributes:
        page_id (Optional[str]):
        page_name (Optional[str]):
    """

    page_id: Optional[str]
    page_name: Optional[str]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.page_id: Optional[str] = None
        self.page_name: Optional[str] = None


@dataclass
class WikiBody(Annotation):

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)


@dataclass
class WikiSection(Annotation):
    """
    Attributes:
        is_intro (Optional[bool]):
    """

    is_intro: Optional[bool]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.is_intro: Optional[bool] = None


@dataclass
class WikiParagraph(Annotation):

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)


@dataclass
class WikiTitle(Annotation):
    """
    Title for the wikipedia sections.
    """

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)


@dataclass
class WikiArticleTitle(Annotation):
    """
    The title for the whole article.
    """

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)


@dataclass
class WikiAnchor(Annotation):
    """
    Attributes:
        target_page_name (Optional[str]):
        is_external (Optional[bool]):	whether this is a link to external (i.e. outside wikipedia).
    """

    target_page_name: Optional[str]
    is_external: Optional[bool]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.target_page_name: Optional[str] = None
        self.is_external: Optional[bool] = None


@dataclass
class WikiInfoBoxProperty(Generics):
    """
    Attributes:
        key (Optional[str]):
        value (Optional[str]):
    """

    key: Optional[str]
    value: Optional[str]

    def __init__(self, pack: DataPack):
        super().__init__(pack)
        self.key: Optional[str] = None
        self.value: Optional[str] = None


@dataclass
class WikiInfoBoxMapped(Generics):
    """
    Attributes:
        key (Optional[str]):
        value (Optional[str]):
    """

    key: Optional[str]
    value: Optional[str]

    def __init__(self, pack: DataPack):
        super().__init__(pack)
        self.key: Optional[str] = None
        self.value: Optional[str] = None


@dataclass
class WikiCategory(Generics):
    """
    Attributes:
        values (List[str]):
    """

    values: List[str]

    def __init__(self, pack: DataPack):
        super().__init__(pack)
        self.values: List[str] = []



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging

from pathlib import Path
from typing import Dict, List, Set, Union, Iterator, Optional, Type, Any, Tuple

import jsonpickle

from sortedcontainers import SortedList
from packaging.version import Version

from forte.common import ProcessExecutionException
from forte.data.base_pack import BaseMeta, BasePack
from forte.data.data_pack import DataPack
from forte.data.index import BaseIndex
from forte.data.ontology.core import Entry
from forte.data.ontology.core import EntryType
from forte.data.ontology.top import (
    Annotation,
    MultiPackGroup,
    MultiPackLink,
    MultiPackEntries,
    MultiPackGeneric,
)
from forte.data.types import DataRequest
from forte.utils import get_class
from forte.version import DEFAULT_PACK_VERSION, PACK_ID_COMPATIBLE_VERSION


logger = logging.getLogger(__name__)

__all__ = [
    "MultiPackMeta",
    "MultiPack",
    "MultiPackLink",
    "MultiPackGroup",
]

MdRequest = Dict[Type[Union[MultiPackLink, MultiPackGroup]], Union[Dict, List]]


class MultiPackMeta(BaseMeta):
    r"""Meta information of a MultiPack."""
    pass


# pylint: disable=too-many-public-methods


class MultiPack(BasePack[Entry, MultiPackLink, MultiPackGroup]):
    r"""A :class:`~forte.data.multi_pack.MultiPack` contains multiple
    `DataPacks` and a collection of
    cross-pack entries (such as links and groups)
    """

    def __init__(self, pack_name: Optional[str] = None):
        super().__init__(pack_name)

        # Store the pack ids of the subpacks. Note that these are UUIDs so
        # they should be globally non-conflicting.
        self._pack_ref: List[int] = []
        # Store the reverse mapping from pack id to the pack index.
        self._inverse_pack_ref: Dict[int, int] = {}

        # Store the pack names.
        self._pack_names: List[str] = []
        # Store the reverse mapping from name to the pack index.
        self._name_index: Dict[str, int] = {}

        # Reference to the real packs.
        self._packs: List[DataPack] = []

        self.links: SortedList[MultiPackLink] = SortedList()
        self.groups: SortedList[MultiPackGroup] = SortedList()
        self.generics: SortedList[MultiPackGeneric] = SortedList()

        # Used to automatically give name to sub packs.
        self.__default_pack_prefix = "_pack"

        self._index: MultiIndex = MultiIndex()

    def __setstate__(self, state):
        r"""In deserialization, we set up the index and the references to the
        data packs inside.
        """
        super().__setstate__(state)

        self.links = SortedList(self.links)
        self.groups = SortedList(self.groups)
        self.generics = SortedList(self.generics)

        self._index = MultiIndex()
        # TODO: index those pointers?
        self._index.update_basic_index(list(self.links))
        self._index.update_basic_index(list(self.groups))
        self._index.update_basic_index(list(self.generics))

        for a in self.links:
            a.set_pack(self)

        for a in self.groups:
            a.set_pack(self)

        for a in self.generics:
            a.set_pack(self)

        # Rebuild the name to index lookup.
        self._name_index = {n: i for (i, n) in enumerate(self._pack_names)}

        # Create the pack list for adding them back.
        self._packs = []

    def relink(self, packs: Iterator[DataPack]):
        """
        Re-link the reference of the multi-pack to other entries, including
        the data packs in it.

        Args:
            packs: a data pack iterator.

        Returns:
            None
        """
        self._packs.extend(packs)
        for a in self.links:
            a.relink_pointer()

        for a in self.groups:
            a.relink_pointer()

        for a in self.generics:
            a.relink_pointer()

    def __getstate__(self):
        r"""
        Pop some recoverable information in serialization.

        Returns:

        """
        state = super().__getstate__()
        # Do not directly serialize the pack itself.
        state.pop("_packs")

        state["links"] = list(state["links"])
        state["groups"] = list(state["groups"])
        state["generics"] = list(state["generics"])

        return state

    def __iter__(self):
        yield from self.links
        yield from self.groups
        yield from self.generics

    def _init_meta(self, pack_name: Optional[str] = None) -> MultiPackMeta:
        return MultiPackMeta(pack_name)

    def _validate(self, entry: EntryType) -> bool:
        return isinstance(entry, MultiPackEntries)

    # TODO: get_subentry maybe useless
    def get_subentry(self, pack_idx: int, entry_id: int):
        r"""
        Get sub_entry from multi pack. This method uses `pack_id` (a unique
        identifier assigned to datapack) to get a pack from multi pack,
        and then return its sub_entry with entry_id. Noted this is changed from
        the way of accessing such pack before the PACK_ID_COMPATIBLE_VERSION,
        in which the `pack_idx` was used as list index number to access/reference
        a pack within the multi pack (and in this case then get the sub_entry).

        Args:
            pack_idx: The pack_id for the data_pack in the
              multi pack.
            entry_id: the id for the entry from the pack with pack_id

        Returns:
            sub-entry of the pack with id = `pack_idx`

        """
        pack_array_index: int = pack_idx  # the old way
        # the following check if the pack version is higher than the (backward)
        # compatible version in which pack_idx is the pack_id not list index
        if Version(self.pack_version) >= Version(PACK_ID_COMPATIBLE_VERSION):
            pack_array_index = self.get_pack_index(
                pack_idx
            )  # the new way: using pack_id instead of array index

        return self._packs[pack_array_index].get_entry(entry_id)
        # return self.get_pack_at(pack_idx).get_entry(entry_id) #old version

    def get_span_text(self, begin: int, end: int):
        raise ValueError(
            "MultiPack objects do not contain text, please refer to a "
            "specific data pack to get text."
        )

    def remove_pack(
        self,
        index_of_pack: int,
        clean_invalid_entries: bool = False,
        purge_lists: bool = False,
    ) -> bool:
        """
        Remove a data pack at index `index_of_pack` from this multi pack.

        In a multi pack, the data pack to be removed may be associated with
        some multi pack entries, such as `MultiPackLinks` that are connected
        with other packs. These entries will become dangling and invalid,
        thus need to be removed. One can consider removing these links before
        calling this function, or set the `clean_invalid_entries` to `True` so
        that they will be automatically pruned. The purge of the lists in this
        multi_pack can be called if `pruge_lists` is set to true which will
        remove the empty spaces in the lists of this multi pack of the removed
        pack and resulting in the index for the remaining packs after the
        removed pack to be changed, so user will be responsible to manage such
        changes if the index(es) of said remaining pack is used or stored
        somewhere by user, after purging the lists.

        Args:
            index_of_pack: The index of pack for removal from the
              multi pack. If invalid, no pack will be deleted.
            clean_invalid_entries: Switch for automatically cleaning
              the entries associated with the data pack being deleted which
              will become invalid after the removal of the pack. Default is
              False.
            purge_lists: Switch for automatically removing the empty
              spaces in the lists of this multi pack of the removed pack and
              resulting in the index for the remaining packs after the removed
              pack to be changed, so user will be responsible to manage such
              changes if the index(es) of said remaining pack is used or stored
              somewhere by user, after purging the lists. Default is False.

        Returns:
            True if successful.

        Raises:
            ValueError: if ``clean_invalid_entries`` is set to False and the
                DataPack to be removed have entries (in links, groups)
                associated with it.

        """
        pack = self.get_pack_at(index_of_pack)

        if pack is None or (not isinstance(pack, DataPack)):
            type_name = "None"
            if pack is not None:
                type_name = type(pack)
            raise ValueError(
                f"Object for the index should be pack, but got "
                f"type: {type_name}"
            )

        return self._remove_pack(
            pack, index_of_pack, clean_invalid_entries, purge_lists
        )

    def _remove_pack(
        self,
        pack: DataPack,
        index_of_pack: int,
        clean_invalid_entries: bool = False,
        purge_lists: bool = False,
    ) -> bool:
        """
        Remove an existing data pack in the multi pack. To prevent index of the
        packs following it being changed, set this empty position to None
        in order to keep the index for the packs intact. in `_pack_ref[]` and
        `_packs[]` the position will be set to None while in `_pack_names[]`
        the position will be set as empty string. The purge of the lists in this
        multi_pack can be called if `pruge_lists` is set to true which will
        remove the empty spaces in the lists of this multi pack of the removed
        pack and resulting in the index for the remaining packs after the
        removed pack to be changed, so user will be responsible to manage such
        changes if the index(es) of said remaining pack is used or stored
        somewhere by user, after purging the lists.

        Args:
            pack: The existing data pack.
            index_of_pack: the index of the pack to be removed
            clean_invalid_entries: Switch for automatically
              cleaning the entries associated with the data pack
              being deleted which will become invalid after the
              removal of the pack. Default is False.
            purge_lists: Switch for automatically removing the empty
              spaces in the lists of this multi pack of the removed pack and
              resulting in the index for the remaining packs after the removed
              pack to be changed, so user will be responsible to manage such
              changes if the index(es) of said remaining pack is used or stored
              somewhere by user, after purging the lists. Default is False.

        Returns:
            True if successful

        Raises:
            ValueError: if ``clean_invalid_entries`` is set to False and the
                DataPack to be removed have entries (in links, groups)
                associated with it.
        """

        # check if the pack to be removed has any cross pack links/groups
        links_with_pack_for_removal = []
        link: MultiPackLink
        for link in self.get(MultiPackLink):
            parent_entry_pid = link.get_parent().pack_id
            child_entry_pid = link.get_child().pack_id
            if (
                parent_entry_pid == pack.pack_id
                or child_entry_pid == pack.pack_id
            ):
                links_with_pack_for_removal.append(link)

        groups_with_pack_for_removal = []
        g: MultiPackGroup
        for g in self.get(MultiPackGroup):
            # e: Annotation
            for e in g.get_members():
                if e.pack_id == pack.pack_id:
                    groups_with_pack_for_removal.append(g)

        if (
            len(links_with_pack_for_removal) > 0
            or len(groups_with_pack_for_removal) > 0
        ):
            if clean_invalid_entries:
                # clean links and groups
                for link in links_with_pack_for_removal:
                    # delete_entry will take care of related indexes
                    self.delete_entry(link)
                for g in groups_with_pack_for_removal:
                    # delete_entry will take care of related indexes
                    self.delete_entry(g)
            else:  # raise exception according to requirement
                raise ValueError(
                    "The pack to be removed has cross-pack references."
                    " Please set clean_invalid_entries to be True to "
                    " auto-remove all references to this pack"
                )

        # To keep the remaining element 's index unchanged, set to None in
        # place instead of direct removal

        self._pack_ref[index_of_pack] = None  # type: ignore

        # Remove the reverse mapping from pack id to the pack index.
        self._inverse_pack_ref.pop(pack.pack_id)

        # Remove the reverse mapping from name to the pack index.
        self._name_index.pop(self.pack_names[index_of_pack])

        # Remove the pack names. To keep the remaining element's index
        # unchanged, set to empty instead of direct removal
        self._pack_names[index_of_pack] = ""
        # Remove Reference to the data pack.
        self._packs[index_of_pack] = None  # type: ignore

        if purge_lists:
            self.purge_deleted_packs()  # keep the behavior consistent

        return True

    def purge_deleted_packs(self) -> bool:
        """
        Purge deleted packs from lists previous set to -1, empty or none to keep index unchanged
        Caution: Purging the deleted_packs from lists in multi_pack will remove the empty spaces
        from the lists of this multi_pack after a pack is removed and resulting the indexes of
        the packs after the deleted pack(s) to change, so user will be responsible to manage such
        changes if such index of a pack is used or stored somewhere in user's code after purging.

        Returns:
            True if successful.
        """

        # Remove those None in place and shrink the _pack_ref list.
        # Caution: item index will change
        for index in range(len(self._pack_ref) - 1, 0, -1):
            if self._pack_ref.__getitem__(index) is None:
                self._pack_ref.__delitem__(index)

        # Remove those None in place and shrink the _pack_names list.
        # Caution: item index will change
        for index in range(len(self._pack_names) - 1, 0, -1):
            if not self._pack_names.__getitem__(index):
                self._pack_names.__delitem__(index)

        # Remove those None in place and shrink the _packs list. Caution: item index will change
        for index in range(len(self._packs) - 1, 0, -1):
            if self._packs.__getitem__(index) is None:
                self._packs.__delitem__(index)

        return True

    def add_pack(
        self, ref_name: Optional[str] = None, pack_name: Optional[str] = None
    ) -> DataPack:
        """
        Create a data pack and add it to this multi pack. If `ref_name` is
        provided, it will be used to index the data pack. Otherwise, a default
        name based on the pack id will be created for this data pack. The
        created data pack will be returned.

        Args:
            ref_name: The pack name used to reference this data pack from
              the multi pack. If none, the reference name will not be set.
            pack_name: The pack name of the data pack (itself). If none,
              the name will not be set.

        Returns: The newly created data pack.

        """
        if ref_name in self._name_index:
            raise ValueError(f"The name {ref_name} has already been taken.")
        if ref_name is not None and not isinstance(ref_name, str):
            type_name = "None"
            if ref_name is not None:
                type_name = type(ref_name)
            raise ValueError(
                f"key of the pack should be str, but got type: {type_name}"
            )

        pack: DataPack = DataPack(pack_name=pack_name)
        self.add_pack_(pack, ref_name)
        return pack

    def add_pack_(self, pack: DataPack, ref_name: Optional[str] = None):
        """
        Add a existing data pack to the multi pack.

        Args:
            pack: The existing data pack.
            ref_name: The name to used in this multi pack.

        Returns:
            None
        """
        if ref_name in self._name_index:
            raise ValueError(f"The name {ref_name} has already been taken.")
        if ref_name is not None and not isinstance(ref_name, str):
            raise ValueError(
                f"key of the pack should be str, but got "
                f""
                f"{type(ref_name)}"
            )
        if not isinstance(pack, DataPack):
            raise ValueError(
                f"value of the packs should be DataPack, but "
                f"got {type(pack)}"
            )

        pid = pack.pack_id

        if ref_name is None:
            # Create a default name based on the pack id.
            ref_name = f"{self.__default_pack_prefix}_{pid}"

        # Record the pack's global id and names. Also the reverse lookup map.
        self._pack_ref.append(pid)
        self._inverse_pack_ref[pid] = len(self._pack_ref) - 1

        self._pack_names.append(ref_name)
        self._name_index[ref_name] = len(self._pack_ref) - 1

        self._packs.append(pack)

    def get_pack_at(self, index: int) -> DataPack:
        """
        Get data pack at provided index.

        Args:
            index: The index of the pack.

        Returns:
            The pack at the index.

        """
        # return self._pack_manager.get_from_pool(self._pack_ref[index])
        return self.packs[index]

    def get_pack_index(self, pack_id: int) -> int:
        """
        Get the pack index from the global pack id.

        Args:
            pack_id: The global pack id to find.

        Returns:
            None
        """
        try:
            return self._inverse_pack_ref[pack_id]
        except KeyError as ke:
            raise ProcessExecutionException(
                f"Pack {pack_id} is not in this multi-pack."
            ) from ke

    def get_pack(self, name: str) -> DataPack:
        """
        Get data pack of name.

        Args:
            name: The name of the pack.

        Returns:
            The pack that has that name.
        """
        return self._packs[self._name_index[name]]

    def pack_ids(self) -> List[int]:
        return self._pack_ref

    @property
    def num_pack(self) -> int:
        return len(self._packs)

    @property
    def packs(self) -> List[DataPack]:
        """
        Get the list of Data packs that in the order of added.

        Please do not use this try

        Returns:
            List of data packs contained in this multi-pack.
        """
        return self._packs

    @property
    def pack_names(self) -> List[str]:
        return self._pack_names

    def update_pack(self, named_packs: Dict[str, DataPack]):
        for pack_name, pack in named_packs.items():
            self.add_pack_(pack, pack_name)

    def iter_packs(self) -> Iterator[Tuple[str, DataPack]]:
        for pack_name, pack in zip(self._pack_names, self.packs):
            yield pack_name, pack

    def rename_pack(self, old_name: str, new_name: str):
        r"""Rename the pack to a new name. If the new_name is already taken, a
        ``ValueError`` will be raised. If the old_name is not found, then a
        ``KeyError`` will be raised just as missing value from a dictionary.

        Args:
            old_name: The old name of the pack.
            new_name: The new name to be assigned for the pack.

        Returns:
            None
        """
        if new_name in self._name_index:
            raise ValueError("The new name is already taken.")
        pack_index = self._name_index[old_name]
        self._name_index[new_name] = pack_index
        self._pack_names[pack_index] = new_name

    @property
    def all_links(self) -> Iterator[MultiPackLink]:
        """
        An iterator of all links in this multi pack.

        Returns:
            Iterator of all links, of type
            :class:`~forte.data.ontology.top.MultiPackLink`.

        """
        yield from self.links

    @property
    def num_links(self) -> int:
        """
        Number of groups in this multi pack.

        Returns:
            Number of links.

        """
        return len(self.groups)

    @property
    def all_groups(self) -> Iterator[MultiPackGroup]:
        """
        An iterator of all groups in this multi pack.

        Returns:
            Iterator of all groups, of type
            :class:`~forte.data.ontology.top.MultiPackGroup`.

        """
        yield from self.groups

    @property
    def num_groups(self) -> int:
        """
        Number of groups in this multi pack.

        Returns:
            Number of groups.

        """
        return len(self.groups)

    @property
    def generic_entries(self) -> Iterator[MultiPackGeneric]:
        yield from self.generics

    def add_all_remaining_entries(self, component: Optional[str] = None):
        """
        Calling this function will add the entries that are not added to the
        pack manually.

        Args:
            component: Overwrite the component record with this.

        Returns:
            None
        """
        super().add_all_remaining_entries(component)
        for pack in self.packs:
            pack.add_all_remaining_entries(component)

    def get_data(
        self,
        context_type,
        request: Optional[DataRequest] = None,
        skip_k: int = 0,
    ) -> Iterator[Dict[str, Any]]:
        raise NotImplementedError(
            "We haven't implemented get data for multi pack data yet."
        )

    def get_single_pack_data(
        self,
        pack_index: int,
        context_type: Type[Annotation],
        request: Optional[DataRequest] = None,
        skip_k: int = 0,
    ) -> Iterator[Dict[str, Any]]:
        r"""Get pack data from one of the packs specified by the name. This is
        equivalent to calling the
        :meth:`~forte.data.data_pack.DataPack.get_data` in
        :class:`~forte.data.data_pack.DataPack`.

        Args:
            pack_index: The index of a single pack.
            context_type: The granularity of the data context, which
                could be any Annotation type.
            request: The entry types and fields required.
                The keys of the dict are the required entry types and the
                value should be either a list of field names or a dict.
                If the value is a dict, accepted items includes "fields",
                "component", and "unit". By setting "component" (a list), users
                can specify the components by which the entries are generated.
                If "component" is not specified, will return entries generated
                by all components. By setting "unit" (a string), users can
                specify a unit by which the annotations are indexed.
                Note that for all annotations, "text" and "span" fields are
                given by default; for all links, "child" and "parent"
                fields are given by default.
            skip_k:Will skip the first k instances and generate
                data from the k + 1 instance.

        Returns:
            A data generator, which generates one piece of data (a dict
            containing the required annotations and context).
        """

        yield from self.get_pack_at(pack_index).get_data(
            context_type, request, skip_k
        )

    def get_cross_pack_data(
        self,
        request: MdRequest,
    ):
        r"""
        .. note::

            This function is not finished.

        Get data via the links and groups across data packs. The keys could be
        `MultiPack` entries (i.e. `MultiPackLink` and `MultiPackGroup`). The
        values specifies the detailed entry information to be get. The value
        can be a List of field names, then the return results will contains all
        specified fields.

        One can also call this method with more constraints by providing
        a dictionary, which can contain the following keys:

        - "fields", this specifies the attribute field names to be obtained
        - "unit", this specifies the unit used to index the annotation
        - "component", this specifies a constraint to take only the entries
          created by the specified component.

        The data request logic is similar to that of
        :meth:`~forte.data.data_pack.DataPack.get_data` function
        in :class:`~forte.data.data_pack.DataPack`, but applied on
        `MultiPack` entries.

        Example:

        .. code-block:: python

            requests = {
                MultiPackLink:
                    {
                        "component": ["dummy"],
                        "fields": ["speaker"],
                    },
                base_ontology.Token: ["pos", "sense""],
                base_ontology.EntityMention: {
                    "unit": "Token",
                },
            }
            pack.get_cross_pack_data(requests)

        Args:
            request: A dict containing the data request. The keys are the types
              to be requested, and the fields are the detailed constraints.

        Returns:
            None
        """
        # TODO: Not finished yet
        pass

    def __add_entry_with_check(
        self, entry: EntryType, allow_duplicate: bool = True
    ) -> EntryType:
        r"""Internal method to add an :class:`~forte.data.ontology.core.Entry` object to the
        :class:`~forte.data.multi_pack.MultiPack` object.

        Args:
            - entry: An :class:`~forte.data.ontology.core.Entry` object
              to be added to the datapack.

            - allow_duplicate: Whether we allow duplicate in the datapack.

        Returns:
            The input entry itself
        """
        if isinstance(entry, MultiPackLink):
            target = self.links
        elif isinstance(entry, MultiPackGroup):
            target = self.groups
        elif isinstance(entry, MultiPackGeneric):
            target = self.generics
        else:
            raise ValueError(
                f"Invalid entry type {type(entry)} for Multipack. A valid "
                f"entry should be an instance of MultiPackLink, MultiPackGroup"
                f", or MultiPackGeneric."
            )

        add_new = allow_duplicate or (entry not in target)

        if add_new:
            target.add(entry)

            # TODO: add the pointers?

            # update the data pack index if needed
            self._index.update_basic_index([entry])
            if self._index.link_index_on and isinstance(entry, MultiPackLink):
                self._index.update_link_index([entry])
            if self._index.group_index_on and isinstance(entry, MultiPackGroup):
                self._index.update_group_index([entry])

            self._pending_entries.pop(entry.tid)
            return entry
        else:
            return target[target.index(entry)]

    def get(  # type: ignore
        self,
        entry_type: Union[str, Type[EntryType]],
        components: Optional[Union[str, List[str]]] = None,
        include_sub_type=True,
    ) -> Iterator[EntryType]:
        """Get entries of ``entry_type`` from this multi pack.

        Example:

        .. code-block:: python

            for relation in pack.get(
                                CrossDocEntityRelation,
                                component="relation_creator"
                                ):
                print(relation.get_parent())

        In the above code snippet, we get entries of type
        :class:`~ft.onto.base_ontology.CrossDocEntityRelation` which were
        generated by a component named ``relation_creator``

        Args:
            entry_type: The type of the entries requested.
            components: The component generating the
                entries requested. If `None`, all valid entries generated by
                any component will be returned.
            include_sub_type: whether to return the sub types of the
                queried `entry_type`. True by default.

        Returns:
            An iterator of the entries matching the arguments, following
            the order of entries (first sort by entry comparison, then by
            insertion)

        """
        entry_type_: Type[EntryType]
        if isinstance(entry_type, str):
            entry_type_ = get_class(entry_type)
            if not issubclass(entry_type_, Entry):
                raise AttributeError(
                    f"The specified entry type [{entry_type}] "
                    f"does not correspond to a "
                    f"'forte.data.ontology.core.Entry' class"
                )
        else:
            entry_type_ = entry_type

        entry_iter: Iterator[Entry]

        if not include_sub_type:
            entry_iter = self.get_entries_of(entry_type_)
        elif issubclass(entry_type_, MultiPackLink):
            entry_iter = self.links
        elif issubclass(entry_type_, MultiPackGroup):
            entry_iter = self.groups
        elif issubclass(entry_type_, MultiPackGeneric):
            entry_iter = self.generics
        else:
            raise ValueError(
                f"The entry type: {entry_type_} is not supported by MultiPack."
            )

        all_types: Set[Type]
        if include_sub_type:
            all_types = self._expand_to_sub_types(entry_type_)

        if components is not None:
            if isinstance(components, str):
                components = [components]

        for e in entry_iter:
            # Will check for the type matching if sub types are also requested.
            if include_sub_type and type(e) not in all_types:
                continue

            # Check for the component.
            if components is not None:
                if not self.is_created_by(e, components):
                    continue

            yield e  # type: ignore

    @classmethod
    def deserialize(
        cls,
        data_path: Union[Path, str],
        serialize_method: str = "jsonpickle",
        zip_pack: bool = False,
    ) -> "MultiPack":
        """
        Deserialize a Multi Pack from a string. Note that this will only
        deserialize the native multi pack content, which means the associated
        DataPacks contained in the MultiPack will not be recovered. A
        followed-up step need to be performed to add the data packs back
        to the multi pack.

        This internally calls the
        internal :meth:`~forte.data.base_pack.BasePack._deserialize`
        function from the :class:`~forte.data.base_pack.BasePack`.

        Args:
            data_path: The serialized string of a Multi pack to be deserialized.
            serialize_method: The method used to serialize the data, this
                should be the same as how serialization is done. The current
                options are `jsonpickle` and `pickle`. The default method
                is `jsonpickle`.
            zip_pack: Boolean value indicating whether the input source is
                zipped.

        Returns:
            An data pack object deserialized from the string.
        """
        # pylint: disable=protected-access
        mp: MultiPack = cls._deserialize(data_path, serialize_method, zip_pack)

        # (fix 595) change the dictionary's key after deserialization from str back to int
        mp._inverse_pack_ref = {
            int(k): v for k, v in mp._inverse_pack_ref.items()
        }

        return mp

    @classmethod
    def from_string(cls, data_content: str):
        # pylint: disable=protected-access
        # can not use explict type hint for mp as pylint does not allow type change
        # from base_pack to multi_pack which is problematic so use jsonpickle instead

        mp = jsonpickle.decode(data_content)
        if not hasattr(mp, "pack_version"):
            mp.pack_version = DEFAULT_PACK_VERSION
        # (fix 595) change the dictionary's key after deserialization from str back to int
        mp._inverse_pack_ref = {  # pylint: disable=no-member
            int(k): v
            for k, v in mp._inverse_pack_ref.items()  # pylint: disable=no-member
        }

        return mp

    def _add_entry(self, entry: EntryType) -> EntryType:
        r"""Force add an :class:`forte.data.ontology.core.Entry` object to the
        :class:`~forte.data.multi_pack.MultiPack` object.

        Allow duplicate entries in a datapack.

        Args:
            entry: An :class:`~forte.data.ontology.core.Entry` object
                to be added to the datapack.

        Returns:
            The input entry itself
        """
        return self.__add_entry_with_check(entry, True)

    def delete_entry(self, entry: EntryType):
        r"""Delete an :class:`~forte.data.ontology.core.Entry` object from the
        :class:`~forte.data.multi_pack.MultiPack`.

        Args:
            entry: An :class:`~forte.data.ontology.core.Entry`
                object to be deleted from the pack.

        """
        if isinstance(entry, MultiPackLink):
            target = self.links
        elif isinstance(entry, MultiPackGroup):
            target = self.groups
        elif isinstance(entry, MultiPackGeneric):
            target = self.generics
        else:
            raise ValueError(
                f"Invalid entry type {type(entry)}. A valid entry "
                f"should be an instance of Annotation, Link, or Group."
            )

        begin = 0
        for i, e in enumerate(target[begin:]):
            if e.tid == entry.tid:
                target.pop(i + begin)
                break

        # update basic index
        self._index.remove_entry(entry)

        # set other index invalid
        self._index.turn_link_index_switch(on=False)
        self._index.turn_group_index_switch(on=False)

    @classmethod
    def validate_link(cls, entry: EntryType) -> bool:
        return isinstance(entry, MultiPackLink)

    @classmethod
    def validate_group(cls, entry: EntryType) -> bool:
        return isinstance(entry, MultiPackGroup)

    def view(self):
        return copy.deepcopy(self)


class MultiIndex(BaseIndex):
    pass



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading


class AtomicCounter:
    """An Atomic Counter that is supposed to be thread safe. Code adopted
    from:
    https://gist.github.com/benhoyt/8c8a8d62debe8e5aa5340373f9c509c7

    >>> counter = AtomicCounter()
    >>> counter.value
    0
    >>> counter.increment()
    1
    >>> counter.increment(2)
    3


    >>> counter = AtomicCounter()
    >>> def incrementor():
    ...     for i in range(100000):
    ...         counter.increment()
    >>> threads = []
    >>> for i in range(5):
    ...     thread = threading.Thread(target=incrementor())
    ...     thread.start()
    ...     threads.append(thread)
    >>> for thread in threads:
    ...     thread.join()
    >>> counter.value
    500000

    """

    def __init__(self, initial=0):
        self.value = initial
        self.__lock = threading.Lock()

    def increment(self, amount=1) -> int:
        with self.__lock:
            self.value += amount
            return self.value


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__all__ = ["ProcessManager"]
from collections import deque
from typing import List, Deque

from forte.process_job import ProcessJob, ProcessJobStatus


class ProcessManager:
    r"""A pipeline level manager that manages global processing information,
    such as the current running components. This is an internal class and
    should only be initialized by the system.

    Attributes:
        pipeline_length (int): The length of the current pipeline being
            executed

        _queues (List[Deque[int]]): A list of queues which hold the jobs for
            each processors. The size of this list is equal to pipeline
            length

        _current_queue_index (int): An index indicating which queue to
            read the data from. A value of -1 indicates read from the reader.

        _current_processor_index (int): An index indicating the
            processor that executes the job

        _unprocessed_queue_indices (List[int]): Each element of this list is
            the index of the first UNPROCESSED element in the corresponding
            queue. Length of this list equals the "pipeline_length".

            If `unprocessed_queue_indices = [0, 2]`

                - This means for the 1st queue, the first UNPROCESSED job is at
                  index-0. All elements from indices `[0, len(queue[0]) )` are
                  UNPROCESSED.

                - Similarly, for the 2nd queue, the first UNPROCESSED job is at
                  index-2. All elements from indices `[2, len(queue[1]))` are
                  UNPROCESSED

        _processed_queue_indices (List [int]):  Each element of this list is
            the index of the last PROCESSED element in the corresponding queue.
            Length of this list equals the "pipeline_length".

            If processed_queue_indices = [0, 2]

                - This means for the 1st queue, the last PROCESSED job is at
                  index-0. Only the first element in queue[0] is PROCESSED

                - Similarly, for the 2nd queue, the last PROCESSED job is at
                  index-2. All elements from indices [0, 2] are PROCESSED

        Args:
            pipeline_length: The length of the current pipeline being executed
    """

    def __init__(self, pipeline_length):
        self._pipeline_length: int = pipeline_length
        self._queues: List[Deque[ProcessJob]] = []
        self._current_queue_index: int = -1
        self._current_processor_index: int = 0
        self._unprocessed_queue_indices: List[int] = []
        self._processed_queue_indices: List[int] = []
        self.reset()

    def reset(self):
        self._queues = [deque() for _ in range(self._pipeline_length)]
        self._current_queue_index = -1
        self._current_processor_index = 0
        self._unprocessed_queue_indices = [0] * self._pipeline_length
        self._processed_queue_indices = [-1] * self._pipeline_length

    @property
    def current_processor_index(self) -> int:
        return self._current_processor_index

    @current_processor_index.setter
    def current_processor_index(self, processor_index: int):
        if processor_index >= len(self._queues):
            raise ValueError(
                f"{processor_index} exceeds the pipeline "
                f"range [0, {self.pipeline_length - 1}]"
            )
        self._current_processor_index = processor_index

    @property
    def current_queue_index(self) -> int:
        return self._current_queue_index

    @current_queue_index.setter
    def current_queue_index(self, queue_index: int):
        if queue_index >= len(self._queues):
            raise ValueError(
                f"{queue_index} exceeds the pipeline range "
                f"[0, {self.pipeline_length - 1}]"
            )
        self._current_queue_index = queue_index

    @property
    def unprocessed_queue_indices(self) -> List[int]:
        return self._unprocessed_queue_indices

    @property
    def processed_queue_indices(self) -> List[int]:
        return self._processed_queue_indices

    @property
    def current_queue(self) -> Deque[ProcessJob]:
        return self._queues[self.current_queue_index]

    @property
    def pipeline_length(self) -> int:
        return self._pipeline_length

    def add_to_queue(self, queue_index: int, job: ProcessJob):
        """
        Add a job to a particular queue.

        Args:
            queue_index: The queue that the job is to be added.
            job: The job to be added.

        Returns:
            None
        """
        if queue_index > len(self._queues):
            raise ValueError(
                f"Queue number {queue_index} exceeds queue "
                f"size {len(self._queues)}"
            )
        else:
            # When a job is added to a queue, it will be
            # consider as unprocessed.
            job.set_status(ProcessJobStatus.UNPROCESSED)
            self._queues[queue_index].append(job)

    def exhausted(self) -> bool:
        r"""Returns True only if the last element remaining in the last queue is
        a poison pack."""

        return (
            len(self._queues[self.pipeline_length - 1]) == 1
            and self._queues[self.pipeline_length - 1][0].is_poison
        )



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file implements SubwordExtractor, which is used to extract feature
from the subwords of an entry.
"""
import logging
from typing import Union, Dict, Optional


from forte.common.configuration import Config
from forte.data.data_pack import DataPack
from forte.data.converter.feature import Feature
from forte.data.base_extractor import BaseExtractor
from forte.data.ontology import Annotation
from forte.utils import create_import_error_msg

logger = logging.getLogger(__name__)

__all__ = ["SubwordExtractor"]


class SubwordExtractor(BaseExtractor):
    r"""SubwordExtractor extracts feature from the subword of entry.
    Most of the time, a user will not need to call this class explicitly,
    they will be called by the framework.

    Args:
        config: An instance of `Dict`  :class:`~forte.common.configuration.Config`
    """

    def initialize(self, config: Union[Dict, Config]):
        # pylint: disable=attribute-defined-outside-init
        super().initialize(config=config)

        try:
            from texar.torch.data.tokenizers.bert_tokenizer import (  # pylint:disable=import-outside-toplevel
                BERTTokenizer,
            )
        except ImportError as e:
            raise ImportError(
                create_import_error_msg(
                    "texar-pytorch", "extractor", "SubwordExtractor"
                )
            ) from e

        self.tokenizer = BERTTokenizer(
            pretrained_model_name=self.config.pretrained_model_name,
            cache_dir=None,
            hparams=None,
        )
        predefined_dict = [key for key, _ in self.tokenizer.vocab.items()]
        self.predefined_vocab(predefined_dict)
        if not self.vocab:
            raise AttributeError("Vocabulary is required in SubwordExtractor.")
        self.vocab.mark_special_element(self.tokenizer.vocab["[PAD]"], "PAD")
        self.vocab.mark_special_element(self.tokenizer.vocab["[UNK]"], "UNK")

    @classmethod
    def default_configs(cls):
        r"""Returns a dictionary of default hyper-parameters.

        Here:

        - "`pretrained_model_name`":
          The name of the pretrained bert model. Must be the same
          as used in subword tokenizer.
        - "`subword_class`": the fully qualified name of the class of the
          subword, default is `ft.onto.base_ontology.Subword`.
        """
        config = super().default_configs()
        config.update(
            {
                "pretrained_model_name": None,
                "subword_class": "ft.onto.base_ontology.Subword",
            }
        )
        return config

    def extract(
        self, pack: DataPack, context: Optional[Annotation] = None
    ) -> Feature:
        r"""Extract the subword feature of one instance.

        Args:
            pack: The datapack that contains the current
                instance.
            context: The context is an Annotation entry where
                features will be extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.

        Returns:
            Feature: a feature that contains the extracted data.
        """
        data = []

        subword: Annotation
        for subword in pack.get(self.config.subword_class, context):
            text = subword.text  # type: ignore
            if not subword.is_first_segment:  # type: ignore
                text = "##" + text
            data.append(self.element2repr(text))

        data = (
            [self.element2repr("[CLS]")] + data + [self.element2repr("[SEP]")]
        )

        meta_data = {
            "need_pad": self.vocab.use_pad,  # type: ignore
            "pad_value": self.get_pad_value(),
            "dim": 1,
            "dtype": int,
        }

        return Feature(data=data, metadata=meta_data, vocab=self.vocab)



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
BERT encoder and BERT classifier. BERT classifier takes in fine-tuned BERT
encoder.
"""
import os
from typing import Optional, cast
from forte.utils import create_import_error_msg
from forte.common.configuration import Config

try:
    import torch
    from torch.nn import Parameter
except ImportError as e1:
    raise ImportError(
        create_import_error_msg("torch", "ir", "Information Retrieval supports")
    ) from e1

try:
    from texar.torch.modules.pretrained import PretrainedBERTMixin
    from texar.torch.modules.encoders import BERTEncoder as TxBERTEncoder
    from texar.torch.modules.classifiers import (
        BERTClassifier as TxBERTClassifier,
    )
except ImportError as e1:
    raise ImportError(
        create_import_error_msg("texar-pytorch", "ir", "IR support")
    ) from e1


__all__ = ["BERTEncoder", "BERTClassifier"]


class BERTClassifier(TxBERTClassifier, PretrainedBERTMixin):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        hparams = self.default_hparams()
        hparams.update(kwargs)

        self.load_pretrained_config(hparams=hparams)

        super().__init__(
            pretrained_model_name=self.pretrained_model_name,
            cache_dir=self.cache_dir,
        )

        self.init_pretrained_weights(prefix=self._hparams.prefix)

    def default_hparams(self):
        params = super().default_hparams()
        params.update(self.kwargs)
        return params

    def load_pretrained_config(
        self,
        pretrained_model_name: Optional[str] = None,
        cache_dir: Optional[str] = None,
        hparams=None,
    ):
        r"""Load paths and configurations of the pre-trained model.

        Args:
            pretrained_model_name (optional): A str with the name
                of a pre-trained model to load. If `None`, will use the model
                name in :attr:`hparams`.
            cache_dir (optional): The path to a folder in which the
                fine-tuned model is present.
            hparams (dict or HParams, optional): Hyperparameters. Missing
                hyperparameter will be set to default values. See
                :meth:`default_hparams` for the hyperparameter structure
                and default values.
        """

        self.pretrained_model_name = (
            hparams["pretrained_model_name"]
            if pretrained_model_name is None
            else pretrained_model_name
        )

        rel_dir = hparams["model_dir"] if cache_dir is None else cache_dir
        self.cache_dir = os.path.join(os.path.dirname(__file__), rel_dir)

        if self.pretrained_model_name is None or self.cache_dir is None:
            raise ValueError(
                "Pre-trained model name and directory should"
                "be defined in the fine tuned BERT model."
            )

        self.pretrained_model_dir = os.path.join(
            self.cache_dir, self.pretrained_model_name
        )

        pretrained_model_hparams = self._transform_config(
            self.pretrained_model_name, self.pretrained_model_dir
        )

        super_params = self.default_hparams()
        if "prefix" not in super_params:
            super_params["prefix"] = "_encoder.encoder."
        self._hparams = Config(pretrained_model_hparams, super_params)

    def _init_from_checkpoint(
        self, pretrained_model_name: str, cache_dir: str, **kwargs
    ):
        # pylint: disable=import-outside-toplevel
        try:
            import numpy as np
            import tensorflow as tf
        except ImportError as e:
            raise ImportError(
                "Loading TensorFlow models in PyTorch requires installing "
                "TensorFlow. Please see https://www.tensorflow.org/install/ "
                "for installation instructions."
            ) from e

        py_prefix = "encoder." if "prefix" not in kwargs else kwargs["prefix"]

        global_tensor_map = {
            "bert/embeddings/word_embeddings": "word_embedder._embedding",
            "bert/embeddings/token_type_embeddings": "segment_embedder._embedding",
            "bert/embeddings/position_embeddings": "position_embedder._embedding",
            "bert/embeddings/LayerNorm/beta": "encoder.input_normalizer.bias",
            "bert/embeddings/LayerNorm/gamma": "encoder.input_normalizer.weight",
        }
        layer_tensor_map = {
            "attention/self/key/bias": "self_attns.{}.K_dense.bias",
            "attention/self/query/bias": "self_attns.{}.Q_dense.bias",
            "attention/self/value/bias": "self_attns.{}.V_dense.bias",
            "attention/output/dense/bias": "self_attns.{}.O_dense.bias",
            "attention/output/LayerNorm/gamma": "poswise_layer_norm.{}.weight",
            "attention/output/LayerNorm/beta": "poswise_layer_norm.{}.bias",
            "intermediate/dense/bias": "poswise_networks.{}._layers.0.bias",
            "output/dense/bias": "poswise_networks.{}._layers.2.bias",
            "output/LayerNorm/gamma": "output_layer_norm.{}.weight",
            "output/LayerNorm/beta": "output_layer_norm.{}.bias",
        }
        layer_transpose_map = {
            "attention/self/key/kernel": "self_attns.{}.K_dense.weight",
            "attention/self/query/kernel": "self_attns.{}.Q_dense.weight",
            "attention/self/value/kernel": "self_attns.{}.V_dense.weight",
            "attention/output/dense/kernel": "self_attns.{}.O_dense.weight",
            "intermediate/dense/kernel": "poswise_networks.{}._layers.0.weight",
            "output/dense/kernel": "poswise_networks.{}._layers.2.weight",
        }
        pooler_map = {
            "bert/pooler/dense/bias": "pooler.0.bias",
            "bert/pooler/dense/kernel": "pooler.0.weight",
        }
        output_map = {
            "output_weights": "_logits_layer.weight",
            "output_bias": "_logits_layer.bias",
        }
        tf_path = os.path.abspath(
            os.path.join(cache_dir, self._MODEL2CKPT[pretrained_model_name])
        )

        # Load weights from TF model
        init_vars = tf.train.list_variables(tf_path)
        tfnames, arrays = [], []
        for name, _ in init_vars:
            array = tf.train.load_variable(tf_path, name)
            tfnames.append(name)
            arrays.append(array.squeeze())

        idx = 0
        for name, array in zip(tfnames, arrays):
            if (
                name.startswith("cls")
                or name == "global_step"
                or name.endswith("adam_m")
                or name.endswith("adam_v")
            ):
                # ignore those variables begin with cls
                # ignore 'global_step' variable
                # ignore optimizer state variable
                continue

            if name in global_tensor_map:
                v_name = "_encoder." + global_tensor_map[name]
                pointer = self._name_to_variable(v_name)
                assert pointer.shape == array.shape
                pointer.data = torch.from_numpy(array)
                idx += 1
            elif name in pooler_map:
                v_name = "_encoder." + pooler_map[name]
                pointer = self._name_to_variable(v_name)
                if name.endswith("bias"):
                    assert pointer.shape == array.shape
                    pointer.data = torch.from_numpy(array)
                    idx += 1
                else:
                    array_t = np.transpose(array)
                    assert pointer.shape == array_t.shape
                    pointer.data = torch.from_numpy(array_t)
                    idx += 1
            elif name in output_map:
                v_name = output_map[name]
                pointer = self._name_to_variable(v_name)
                if name.endswith("bias"):
                    assert pointer.shape == array.shape
                    pointer.data = torch.from_numpy(array)
                    idx += 1
                else:
                    assert pointer.shape == array.shape
                    pointer.data = torch.from_numpy(array)
                    idx += 1
            else:
                # here name is the TensorFlow variable name
                name_tmp = name.split("/")
                # e.g. layer_
                layer_no = name_tmp[2][6:]
                name_tmp = "/".join(name_tmp[3:])
                if name_tmp in layer_tensor_map:
                    v_name = layer_tensor_map[name_tmp].format(layer_no)
                    pointer = self._name_to_variable(py_prefix + v_name)
                    assert pointer.shape == array.shape
                    pointer.data = torch.from_numpy(array)
                elif name_tmp in layer_transpose_map:
                    v_name = layer_transpose_map[name_tmp].format(layer_no)
                    pointer = self._name_to_variable(py_prefix + v_name)
                    array_t = np.transpose(array)
                    assert pointer.shape == array_t.shape
                    pointer.data = torch.from_numpy(array_t)
                else:
                    raise NameError(f"Variable with name '{name}' not found")
                idx += 1

    def _name_to_variable(self, name: str) -> Parameter:
        """
        Cast return value of method of the super class
        `texar.torch.module.PretrainedBertMixin._name_to_variable` to avoid
        type errors
        """
        return cast(Parameter, super()._name_to_variable(name))


class BERTEncoder(TxBERTEncoder):
    # pylint: disable=unused-argument
    def load_pretrained_config(
        self,
        pretrained_model_name: Optional[str] = None,
        cache_dir: Optional[str] = None,
        hparams=None,
    ):
        # pylint: disable=attribute-defined-outside-init
        self.pretrained_model_name = pretrained_model_name
        self.cache_dir = cache_dir

    def init_pretrained_weights(self):
        pass

    @staticmethod
    def default_hparams():
        params = BERTEncoder.default_hparams()
        params["model_dir"] = None
        return params



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Dict, Any

from forte.data.data_pack import DataPack
from forte.processors.base import PackProcessor

__all__ = [
    "LowerCaserProcessor",
]


class LowerCaserProcessor(PackProcessor):
    def _process(self, input_pack: DataPack):
        text = input_pack.text
        for src, tgt in self.configs.custom_substitutions:
            text = text.replace(src, tgt)

        lower_text = text.lower()
        if len(lower_text) == len(text):
            input_pack.set_text(lower_text.lower())
        else:
            error_char = ""
            for c in text:
                if not c.lower().upper() == c:
                    error_char = c
                    break

            logging.error(
                "Some characters cannot be converted to lower case without "
                "changing length in pack [%s] will "
                "result in a change of text length, which will cause "
                "problems in the data pack system. The text of this pack "
                "will remain unchanged. One way to solve this is to provide "
                "values from the 'custom_substitutions'. The first "
                "problematic character is [%s].",
                input_pack.pack_id,
                error_char,
            )

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        """
        Default configurations for this processor, it contains the following
        configuration values:

        - "custom_substitutions": a dictionary contains the mapping
            used to conduct lower case, {"İ": "i"}. The length (`len`) of the
            two string must be the same.

        Returns:

        """
        return {
            "custom_substitutions": {},
            "@no_typecheck": ["custom_substitutions"],
        }



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Data augmentation processors from the paper "EDA: Easy Data Augmentation
Techniques for Boosting Performance on Text Classification Tasks", including
Random Swap, Random Insertion and Random Deletion. All three processors are
implemented based on the ReplacementDataAugmentProcessor.
"""

from math import ceil
import random
from typing import List, Dict, Iterable, Union, Any
from forte.common.configuration import Config
from forte.data.data_pack import DataPack
from forte.data.ontology import Annotation
from forte.processors.data_augment.algorithms.base_data_augmentation_op import (
    BaseDataAugmentationOp,
)
from forte.utils.utils import get_class, create_class_with_kwargs

__all__ = [
    "RandomSwapDataAugmentOp",
    "RandomInsertionDataAugmentOp",
    "RandomDeletionDataAugmentOp",
]

english_stopwords = [
    "i",
    "me",
    "my",
    "myself",
    "we",
    "our",
    "ours",
    "ourselves",
    "you",
    "you're",
    "you've",
    "you'll",
    "you'd",
    "your",
    "yours",
    "yourself",
    "yourselves",
    "he",
    "him",
    "his",
    "himself",
    "she",
    "she's",
    "her",
    "hers",
    "herself",
    "it",
    "it's",
    "its",
    "itself",
    "they",
    "them",
    "their",
    "theirs",
    "themselves",
    "what",
    "which",
    "who",
    "whom",
    "this",
    "that",
    "that'll",
    "these",
    "those",
    "am",
    "is",
    "are",
    "was",
    "were",
    "be",
    "been",
    "being",
    "have",
    "has",
    "had",
    "having",
    "do",
    "does",
    "did",
    "doing",
    "a",
    "an",
    "the",
    "and",
    "but",
    "if",
    "or",
    "because",
    "as",
    "until",
    "while",
    "of",
    "at",
    "by",
    "for",
    "with",
    "about",
    "against",
    "between",
    "into",
    "through",
    "during",
    "before",
    "after",
    "above",
    "below",
    "to",
    "from",
    "up",
    "down",
    "in",
    "out",
    "on",
    "off",
    "over",
    "under",
    "again",
    "further",
    "then",
    "once",
    "here",
    "there",
    "when",
    "where",
    "why",
    "how",
    "all",
    "any",
    "both",
    "each",
    "few",
    "more",
    "most",
    "other",
    "some",
    "such",
    "no",
    "nor",
    "not",
    "only",
    "own",
    "same",
    "so",
    "than",
    "too",
    "very",
    "s",
    "t",
    "can",
    "will",
    "just",
    "don",
    "don't",
    "should",
    "should've",
    "now",
    "d",
    "ll",
    "m",
    "o",
    "re",
    "ve",
    "y",
    "ain",
    "aren",
    "aren't",
    "couldn",
    "couldn't",
    "didn",
    "didn't",
    "doesn",
    "doesn't",
    "hadn",
    "hadn't",
    "hasn",
    "hasn't",
    "haven",
    "haven't",
    "isn",
    "isn't",
    "ma",
    "mightn",
    "mightn't",
    "mustn",
    "mustn't",
    "needn",
    "needn't",
    "shan",
    "shan't",
    "shouldn",
    "shouldn't",
    "wasn",
    "wasn't",
    "weren",
    "weren't",
    "won",
    "won't",
    "wouldn",
    "wouldn't",
]


class RandomSwapDataAugmentOp(BaseDataAugmentationOp):
    r"""
    Data augmentation operation for the Random Swap operation.
    Randomly choose two words in the sentence and swap their positions.
    Do this n times, where n = alpha * input length.
    """

    def augment(self, data_pack: DataPack) -> bool:
        augment_entry = get_class(self.configs["augment_entry"])
        if not issubclass(augment_entry, Annotation):
            raise ValueError(
                f"This augmenter only accept data of "
                f"'forte.data.ontology.top.Annotation' type, "
                f"but {self.configs['augment_entry']} is not."
            )

        annotations: List[Annotation] = list(
            data_pack.get(self.configs["augment_entry"])
        )
        if len(annotations) > 0:
            replace_map: Dict = {}
        for _ in range(ceil(self.configs["alpha"] * len(annotations))):
            swap_idx = random.sample(range(len(annotations)), 2)
            new_idx_0 = (
                swap_idx[1]
                if swap_idx[1] not in replace_map
                else replace_map[swap_idx[1]]
            )
            new_idx_1 = (
                swap_idx[0]
                if swap_idx[0] not in replace_map
                else replace_map[swap_idx[0]]
            )
            replace_map[swap_idx[0]] = new_idx_0
            replace_map[swap_idx[1]] = new_idx_1

        for idx, replace_target in replace_map.items():
            try:
                self.replace_annotations(
                    annotations[idx], annotations[replace_target].text
                )
            except ValueError:
                return False
        return True

    @classmethod
    def default_configs(cls):
        """
        Additional keys for Random Swap:

            - augment_entry (str):
                Defines the entry the processor will augment.
                It should be a full qualified name of the entry class.
                For example, "ft.onto.base_ontology.Sentence".

            - alpha:
                0 <= alpha <= 1. indicates the percent of the words
                in a sentence that are changed. The processor will perform
                the Random Swap operation (input length * alpha) times.
                Default Value is 0.1.

        Returns:
            A dictionary with the default config for this processor.
        """
        return {
            "augment_entry": "ft.onto.base_ontology.Token",
            "other_entry_policy": {
                "ft.onto.base_ontology.Document": "auto_align",
                "ft.onto.base_ontology.Sentence": "auto_align",
            },
            "alpha": 0.1,
        }


class RandomInsertionDataAugmentOp(BaseDataAugmentationOp):
    r"""
    Data augmentation operation for the Random Insertion operation.
    Find a random synonym of a random word in the sentence that is
    not a stop word. Insert that synonym into a random position in
    the sentence. Do this n times, where n = alpha * input length.
    """

    def __init__(self, configs: Union[Config, Dict[str, Any]]) -> None:
        super().__init__(configs)
        self.stopwords = set(self.configs["stopwords"])

    def augment(self, data_pack: DataPack) -> bool:

        replacement_op = create_class_with_kwargs(
            self.configs["insertion_op_configs"]["type"],
            class_args={
                "configs": self.configs["insertion_op_configs"]["kwargs"]
            },
        )

        annotations: List[Annotation] = []
        pos = [0]
        annos: Iterable[Annotation] = data_pack.get(
            self.configs["augment_entry"]
        )
        for anno in annos:
            if anno.text not in self.stopwords:
                annotations.append(anno)
                pos.append(anno.end)
        if len(annotations) > 0:
            for _ in range(ceil(self.configs["alpha"] * len(annotations))):
                src_anno = random.choice(annotations)
                try:
                    _, replaced_text = replacement_op.single_annotation_augment(
                        src_anno
                    )
                except ValueError:
                    return False

                insert_pos = random.choice(pos)
                if insert_pos > 0:
                    replaced_text = " " + replaced_text
                else:
                    replaced_text = replaced_text + " "

                try:
                    self.insert_annotated_span(
                        replaced_text,
                        data_pack,
                        insert_pos,
                        self.configs["augment_entry"],
                    )
                except ValueError:
                    return False
        return True

    @classmethod
    def default_configs(cls):
        """
        Additional keys for Random Swap:

        - augment_entry (str):
            Defines the entry the processor will augment.
            It should be a full qualified name of the entry class.
            For example, "ft.onto.base_ontology.Sentence".

        - alpha:
            0 <= alpha <= 1. indicates the percent of the words
            in a sentence that are changed. The processor will perform
            the Random Insertion operation (input length * alpha) times.
            Default Value is 0.1

        - stopwords:
            a list of stopword for the language.

        - `insertion_op_config`:
            A dictionary representing the configurations
            required operation to take random annotations
            from the source data pack, augment them based
            on specified rules and insert them in random
            positions.

            - type:
                The type of data augmentation operation to be used
                (pass the path of the class which defines the
                required operation)

            - kwargs:
                This dictionary contains the data that is to be
                fed to the required operation (Make sure to be
                well versed with the required configurations of the
                operation that is defined in the type config).

            .. code-block:: python

                {
                    "type": "forte.processors.data_augment.algorithms."
                    "dictionary_replacement_op.DictionaryReplacementOp",
                    "kwargs":{
                        "dictionary_class": (
                            "forte.processors.data_augment."
                            "algorithms.dictionary.WordnetDictionary"
                        ),
                        "prob": 1.0,
                        "lang": "eng",
                    }
                }

        Returns:
            A dictionary with the default config for this processor.
            By default, we use Dictionary Replacement with Wordnet to get
            synonyms to insert.
        """
        return {
            "augment_entry": "ft.onto.base_ontology.Token",
            "other_entry_policy": {
                "ft.onto.base_ontology.Document": "auto_align",
                "ft.onto.base_ontology.Sentence": "auto_align",
            },
            "insertion_op_configs": {
                "type": "forte.processors.data_augment.algorithms."
                "dictionary_replacement_op.DictionaryReplacementOp",
                "kwargs": {
                    "dictionary_class": (
                        "forte.processors.data_augment."
                        "algorithms.dictionary.WordnetDictionary"
                    ),
                    "prob": 1.0,
                    "lang": "eng",
                },
            },
            "alpha": 0.1,
            "stopwords": english_stopwords,
        }


class RandomDeletionDataAugmentOp(BaseDataAugmentationOp):
    r"""
    Data augmentation operation for the Random Insertion operation.
    Randomly remove each word in the sentence with probability alpha.
    """

    def augment(self, data_pack: DataPack) -> bool:
        anno: Annotation
        for anno in data_pack.get(self.configs["augment_entry"]):
            if random.random() < self.configs["alpha"]:
                try:
                    self.delete_annotation(anno)
                except ValueError:
                    return False
        return True

    @classmethod
    def default_configs(cls):
        """
        Returns:
            A dictionary with the default config for this processor.
            Additional keys for Random Deletion:

            - augment_entry (str):
                Defines the entry the processor will augment.
                It should be a full qualified name of the entry class.
                For example, "ft.onto.base_ontology.Sentence".
                Default Value is 0.1

            - alpha:
                0 <= alpha <= 1. The probability to delete each word.
        """
        return {
            "augment_entry": "ft.onto.base_ontology.Token",
            "other_entry_policy": {
                "ft.onto.base_ontology.Document": "auto_align",
                "ft.onto.base_ontology.Sentence": "auto_align",
            },
            "alpha": 0.1,
        }



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict

from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.processors.base import PackProcessor

__all__ = ["AttributeMasker"]

from forte.utils import get_class


class AttributeMasker(PackProcessor):
    def __init__(self):
        super().__init__()
        self.fields: Dict = {}

    # pylint: disable=attribute-defined-outside-init
    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        for entry_type, entry_attributes in self.configs.requests.items():
            entry_class = get_class(entry_type)
            self.fields[entry_class] = entry_attributes

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        r"""Default config for this processor.

        Example usage is shown below

        .. code-block:: python

            {
                "requests": {
                    "ft.onto.base_ontology.Token": ["pos"]
                }
            }

        Here:

        - `"requests"`: dict
          The entry types and fields required. The keys of the requests dict
          are the entry types whose fields need to be masked and the value is
          a list of field names.

        """
        return {"requests": {}, "@no_typecheck": {"requests"}}

    def _process(self, input_pack: DataPack):
        for entry_type, attributes in self.fields.items():
            for entry in input_pack.get_entries_of(entry_type):
                for attribute in attributes:
                    setattr(entry, attribute, None)



# ***automatically_generated***
# ***source json:forte/ontology_specs/base_ontology.json***
# flake8: noqa
# mypy: ignore-errors
# pylint: skip-file
"""
Automatically generated ontology base_ontology. Do not change manually.
"""

from dataclasses import dataclass
from forte.data.base_pack import PackType
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.data.ontology.core import Entry
from forte.data.ontology.core import FDict
from forte.data.ontology.core import FList
from forte.data.ontology.top import Annotation
from forte.data.ontology.top import AudioAnnotation
from forte.data.ontology.top import Generics
from forte.data.ontology.top import Group
from forte.data.ontology.top import Link
from forte.data.ontology.top import MultiPackLink
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional

__all__ = [
    "Token",
    "Subword",
    "Classification",
    "Document",
    "Sentence",
    "Phrase",
    "UtteranceContext",
    "Utterance",
    "PredicateArgument",
    "EntityMention",
    "EventMention",
    "PredicateMention",
    "PredicateLink",
    "Dependency",
    "EnhancedDependency",
    "RelationLink",
    "CrossDocEntityRelation",
    "CoreferenceGroup",
    "EventRelation",
    "CrossDocEventRelation",
    "ConstituentNode",
    "Title",
    "Body",
    "MCOption",
    "MCQuestion",
    "MRCQuestion",
    "Recording",
    "AudioUtterance",
]


@dataclass
class Token(Annotation):
    """
    A span based annotation :class:`Token`, used to represent a token or a word.
    Attributes:
        pos (Optional[str]):
        ud_xpos (Optional[str]):	Language specific pos tag. Used in CoNLL-U Format. Refer to https://universaldependencies.org/format.html
        lemma (Optional[str]):	Lemma or stem of word form.
        chunk (Optional[str]):
        ner (Optional[str]):
        sense (Optional[str]):
        is_root (Optional[bool]):
        ud_features (Dict[str, str]):
        ud_misc (Dict[str, str]):
    """

    pos: Optional[str]
    ud_xpos: Optional[str]
    lemma: Optional[str]
    chunk: Optional[str]
    ner: Optional[str]
    sense: Optional[str]
    is_root: Optional[bool]
    ud_features: Dict[str, str]
    ud_misc: Dict[str, str]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.pos: Optional[str] = None
        self.ud_xpos: Optional[str] = None
        self.lemma: Optional[str] = None
        self.chunk: Optional[str] = None
        self.ner: Optional[str] = None
        self.sense: Optional[str] = None
        self.is_root: Optional[bool] = None
        self.ud_features: Dict[str, str] = dict()
        self.ud_misc: Dict[str, str] = dict()


@dataclass
class Subword(Annotation):
    """
    Used to represent subword tokenization results.
    Attributes:
        is_first_segment (Optional[bool]):
        is_unk (Optional[bool]):
        vocab_id (Optional[int]):
    """

    is_first_segment: Optional[bool]
    is_unk: Optional[bool]
    vocab_id: Optional[int]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.is_first_segment: Optional[bool] = None
        self.is_unk: Optional[bool] = None
        self.vocab_id: Optional[int] = None


@dataclass
class Classification(Generics):
    """
    Used to store values for classification prediction
    Attributes:
        classification_result (Dict[str, float]):
    """

    classification_result: Dict[str, float]

    def __init__(self, pack: DataPack):
        super().__init__(pack)
        self.classification_result: Dict[str, float] = dict()


@dataclass
class Document(Annotation):
    """
    A span based annotation `Document`, normally used to represent a document.
    Attributes:
        document_class (List[str]):	A list of class names that the document belongs to.
        sentiment (Dict[str, float]):
        classifications (FDict[str, Classification]):	Stores the classification results for this document. The key is the name/task of the classification, the value is an classification object storing the results.
    """

    document_class: List[str]
    sentiment: Dict[str, float]
    classifications: FDict[str, Classification]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.document_class: List[str] = []
        self.sentiment: Dict[str, float] = dict()
        self.classifications: FDict[str, Classification] = FDict(self)


@dataclass
class Sentence(Annotation):
    """
    A span based annotation `Sentence`, normally used to represent a sentence.
    Attributes:
        speaker (Optional[str]):
        part_id (Optional[int]):
        sentiment (Dict[str, float]):
        classification (Dict[str, float]):
        classifications (FDict[str, Classification]):	Stores the classification results for this sentence. The key is the name/task of the classification, the value is an classification object storing the results.
    """

    speaker: Optional[str]
    part_id: Optional[int]
    sentiment: Dict[str, float]
    classification: Dict[str, float]
    classifications: FDict[str, Classification]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.speaker: Optional[str] = None
        self.part_id: Optional[int] = None
        self.sentiment: Dict[str, float] = dict()
        self.classification: Dict[str, float] = dict()
        self.classifications: FDict[str, Classification] = FDict(self)


@dataclass
class Phrase(Annotation):
    """
    A span based annotation `Phrase`.
    Attributes:
        phrase_type (Optional[str]):
        headword (Optional[Token]):
    """

    phrase_type: Optional[str]
    headword: Optional[Token]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.phrase_type: Optional[str] = None
        self.headword: Optional[Token] = None


@dataclass
class UtteranceContext(Annotation):
    """
    `UtteranceContext` represents the context part in dialogue.
    """

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)


@dataclass
class Utterance(Annotation):
    """
    A span based annotation `Utterance`, normally used to represent an utterance in dialogue.
    Attributes:
        speaker (Optional[str]):
    """

    speaker: Optional[str]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.speaker: Optional[str] = None


@dataclass
class PredicateArgument(Annotation):
    """
    A span based annotation `PredicateArgument`, normally used to represent an argument of a predicate, can be linked to the predicate via the predicate link.
    Attributes:
        ner_type (Optional[str]):
        predicate_lemma (Optional[str]):
        is_verb (Optional[bool]):
    """

    ner_type: Optional[str]
    predicate_lemma: Optional[str]
    is_verb: Optional[bool]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.ner_type: Optional[str] = None
        self.predicate_lemma: Optional[str] = None
        self.is_verb: Optional[bool] = None


@dataclass
class EntityMention(Annotation):
    """
    A span based annotation `EntityMention`, normally used to represent an Entity Mention in a piece of text.
    Attributes:
        ner_type (Optional[str]):
    """

    ner_type: Optional[str]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.ner_type: Optional[str] = None


@dataclass
class EventMention(Annotation):
    """
    A span based annotation `EventMention`, used to refer to a mention of an event.
    Attributes:
        event_type (Optional[str]):
    """

    event_type: Optional[str]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.event_type: Optional[str] = None


@dataclass
class PredicateMention(Phrase):
    """
    A span based annotation `PredicateMention`, normally used to represent a predicate (normally verbs) in a piece of text.
    Attributes:
        predicate_lemma (Optional[str]):
        framenet_id (Optional[str]):
        is_verb (Optional[bool]):
    """

    predicate_lemma: Optional[str]
    framenet_id: Optional[str]
    is_verb: Optional[bool]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.predicate_lemma: Optional[str] = None
        self.framenet_id: Optional[str] = None
        self.is_verb: Optional[bool] = None


@dataclass
class PredicateLink(Link):
    """
    A `Link` type entry which represent a semantic role link between a predicate and its argument.
    Attributes:
        arg_type (Optional[str]):	The predicate link type.
    """

    arg_type: Optional[str]

    ParentType = PredicateMention
    ChildType = PredicateArgument

    def __init__(self, pack: DataPack, parent: Optional[Entry] = None, child: Optional[Entry] = None):
        super().__init__(pack, parent, child)
        self.arg_type: Optional[str] = None


@dataclass
class Dependency(Link):
    """
    A `Link` type entry which represent a syntactic dependency.
    Attributes:
        dep_label (Optional[str]):	The dependency label.
        rel_type (Optional[str]):
    """

    dep_label: Optional[str]
    rel_type: Optional[str]

    ParentType = Token
    ChildType = Token

    def __init__(self, pack: DataPack, parent: Optional[Entry] = None, child: Optional[Entry] = None):
        super().__init__(pack, parent, child)
        self.dep_label: Optional[str] = None
        self.rel_type: Optional[str] = None


@dataclass
class EnhancedDependency(Link):
    """
    A `Link` type entry which represent a `enhanced dependency
    <https://universaldependencies.org/u/overview/enhanced-syntax.html>`_.

    Attributes:
        dep_label (Optional[str]):	The enhanced dependency label in Universal Dependency.
    """

    dep_label: Optional[str]

    ParentType = Token
    ChildType = Token

    def __init__(self, pack: DataPack, parent: Optional[Entry] = None, child: Optional[Entry] = None):
        super().__init__(pack, parent, child)
        self.dep_label: Optional[str] = None


@dataclass
class RelationLink(Link):
    """
    A `Link` type entry which represent a relation between two entity mentions
    Attributes:
        rel_type (Optional[str]):	The type of the relation.
    """

    rel_type: Optional[str]

    ParentType = EntityMention
    ChildType = EntityMention

    def __init__(self, pack: DataPack, parent: Optional[Entry] = None, child: Optional[Entry] = None):
        super().__init__(pack, parent, child)
        self.rel_type: Optional[str] = None


@dataclass
class CrossDocEntityRelation(MultiPackLink):
    """
    A `Link` type entry which represent a relation between two entity mentions across the packs.
    Attributes:
        rel_type (Optional[str]):	The type of the relation.
    """

    rel_type: Optional[str]

    ParentType = EntityMention
    ChildType = EntityMention

    def __init__(self, pack: MultiPack, parent: Optional[Entry] = None, child: Optional[Entry] = None):
        super().__init__(pack, parent, child)
        self.rel_type: Optional[str] = None


@dataclass
class CoreferenceGroup(Group):
    """
    A group type entry that take `EntityMention`, as members, used to represent coreferent group of entities.
    """

    MemberType = EntityMention

    def __init__(self, pack: DataPack, members: Optional[Iterable[Entry]] = None):
        super().__init__(pack, members)


@dataclass
class EventRelation(Link):
    """
    A `Link` type entry which represent a relation between two event mentions.
    Attributes:
        rel_type (Optional[str]):	The type of the relation.
    """

    rel_type: Optional[str]

    ParentType = EventMention
    ChildType = EventMention

    def __init__(self, pack: DataPack, parent: Optional[Entry] = None, child: Optional[Entry] = None):
        super().__init__(pack, parent, child)
        self.rel_type: Optional[str] = None


@dataclass
class CrossDocEventRelation(MultiPackLink):
    """
    A `Link` type entry which represent a relation between two event mentions across the packs.
    Attributes:
        rel_type (Optional[str]):	The type of the relation.
    """

    rel_type: Optional[str]

    ParentType = EventMention
    ChildType = EventMention

    def __init__(self, pack: MultiPack, parent: Optional[Entry] = None, child: Optional[Entry] = None):
        super().__init__(pack, parent, child)
        self.rel_type: Optional[str] = None


@dataclass
class ConstituentNode(Annotation):
    """
    A span based annotation `ConstituentNode` to represent constituents in constituency parsing. This can also sentiment values annotated on the nodes.
    Attributes:
        label (Optional[str]):
        sentiment (Dict[str, float]):
        is_root (Optional[bool]):
        is_leaf (Optional[bool]):
        parent_node (Optional['ConstituentNode']):
        children_nodes (FList['ConstituentNode']):
    """

    label: Optional[str]
    sentiment: Dict[str, float]
    is_root: Optional[bool]
    is_leaf: Optional[bool]
    parent_node: Optional['ConstituentNode']
    children_nodes: FList['ConstituentNode']

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.label: Optional[str] = None
        self.sentiment: Dict[str, float] = dict()
        self.is_root: Optional[bool] = None
        self.is_leaf: Optional[bool] = None
        self.parent_node: Optional['ConstituentNode'] = None
        self.children_nodes: FList['ConstituentNode'] = FList(self)


@dataclass
class Title(Annotation):
    """
    A span based annotation `Title`, normally used to represent a title.
    """

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)


@dataclass
class Body(Annotation):
    """
    A span based annotation `Body`, normally used to represent a document body.
    """

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)


@dataclass
class MCOption(Annotation):

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)


@dataclass
class MCQuestion(Annotation):
    """
    Attributes:
        options (FList[MCOption]):
        answers (List[int]):
    """

    options: FList[MCOption]
    answers: List[int]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.options: FList[MCOption] = FList(self)
        self.answers: List[int] = []


@dataclass
class MRCQuestion(Annotation):
    """
    An `Annotation` type which represents an MRC question.
    Attributes:
        qid (Optional[int]):
        answers (FList[Phrase]):
    """

    qid: Optional[int]
    answers: FList[Phrase]

    def __init__(self, pack: DataPack, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.qid: Optional[int] = None
        self.answers: FList[Phrase] = FList(self)


@dataclass
class Recording(AudioAnnotation):
    """
    A span based annotation `Recording`, normally used to represent a recording.
    Attributes:
        recording_class (List[str]):	A list of class names that the recording belongs to.
    """

    recording_class: List[str]

    def __init__(self, pack: PackType, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.recording_class: List[str] = []


@dataclass
class AudioUtterance(AudioAnnotation):
    """
    A span based annotation `AudioUtterance`, normally used to represent an utterance in dialogue.
    Attributes:
        speaker (Optional[str]):
    """

    speaker: Optional[str]

    def __init__(self, pack: PackType, begin: int, end: int):
        super().__init__(pack, begin, end)
        self.speaker: Optional[str] = None



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This defines the switcher interface, which can cast the pack types in the
middle of a pipeline flow. The main difference of this from Selector is that
the returned pack will be used to replaced the original pack, while in Selector,
the original pack is kept unchanged.
"""
from abc import ABC
from typing import Generic, TypeVar

from forte.data.base_pack import BasePack
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.pipeline_component import PipelineComponent

InputPackType = TypeVar("InputPackType", bound=BasePack)
OutputPackType = TypeVar("OutputPackType", bound=BasePack)

__all__ = ["Caster", "MultiPackBoxer", "MultiPackUnboxer"]


class Caster(
    PipelineComponent[InputPackType],
    Generic[InputPackType, OutputPackType],
    ABC,
):
    def cast(self, pack: InputPackType) -> OutputPackType:
        raise NotImplementedError

    @staticmethod
    def input_pack_type():
        raise NotImplementedError

    @staticmethod
    def output_pack_type():
        raise NotImplementedError


class MultiPackBoxer(Caster[DataPack, MultiPack]):
    """
    This class creates a MultiPack from a DataPack, this MultiPack will only
    contains the original DataPack, indexed by the :attr:`pack_name`.
    """

    def cast(self, pack: DataPack) -> MultiPack:
        """
        Auto-box the DataPack into a MultiPack by simple wrapping.

        Args:
            pack: The DataPack to be boxed

        Returns:
            An iterator that produces the boxed MultiPack.

        """
        # p = MultiPack()
        pack_name = pack.pack_name + "_multi" if pack.pack_name else None
        # if pack_name in p._name_index:
        #     raise ValueError(f"The name {pack_name} has already been taken.")
        p = MultiPack(pack_name=pack_name)
        p.add_pack_(pack, self.configs.pack_name)
        return p

    @classmethod
    def default_configs(cls):
        return {"pack_name": "default"}

    @staticmethod
    def input_pack_type():
        return DataPack

    @staticmethod
    def output_pack_type():
        return MultiPack


class MultiPackUnboxer(Caster[MultiPack, DataPack]):
    """
    This passes on a single DataPack within the MultiPack.
    """

    def cast(self, pack: MultiPack) -> DataPack:
        """
        Auto-box the MultiPack into a DataPack by using pack_index to take the unique pack.

        Args:
            pack: The MultiPack to be boxed.

        Returns:
            A DataPack boxed from the MultiPack.

        """

        if self.configs.pack_index < pack.num_pack:
            p = pack.get_pack_at(self.configs.pack_index)
            return p
        else:
            raise IndexError(
                f"pack_index: {self.configs.pack_index} is not in this multi-pack."
            )

    @classmethod
    def default_configs(cls):
        return {"pack_index": 0}

    @staticmethod
    def input_pack_type():
        return MultiPack

    @staticmethod
    def output_pack_type():
        return DataPack



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file implements the UDA(Unsupervised Data Augmentation)
algorithm as described in
Unsupervised Data Augmentation for Consistency Training
(https://arxiv.org/abs/1904.12848)
(https://github.com/google-research/uda)
"""
from typing import Optional, List, Tuple, Iterator
from torch import Tensor
from forte.utils import create_import_error_msg

try:
    from texar.torch.losses.info_loss import kl_divg_loss_with_logits
    from texar.torch.data import DataIterator
    from texar.torch.data.data.dataset_utils import Batch
except ImportError as err:
    raise ImportError(
        create_import_error_msg(
            "texar-pytorch", "models", "Unsupervised Data Augmentation(UDA)"
        )
    ) from err


__all__ = [
    "UDAIterator",
]


class UDAIterator:
    r"""
    This iterator wraps the Unsupervised Data Augmentation(UDA)
    algorithm by calculating the unsupervised loss automatically
    during each iteration. It takes both supervised and unsupervised
    data iterator as input.

    The unsupervised data should contain the original input and the
    augmented input. The original and augmented inputs should be in
    the same training example.

    During each iteration, the iterator will return the supervised and
    unsupervised batches. Users can call the :func:`calculate_uda_loss`
    to get the UDA loss and combine it with the supervised loss
    for model training.

    It uses tricks such as prediction sharpening and confidence masking.
    Please refer to the UDA paper for more details.
    (https://arxiv.org/abs/1904.12848)

    Args:
        sup_iterator: The iterator for supervised data. Each item is a
            training/evaluation/test example with key-value pairs as inputs.
        unsup_iterator: The iterator for unsupervised data. Each training
            example in it should contain both the original and augmented data.
        softmax_temperature: The softmax temperature for sharpening the
            distribution. The value should be larger than 0. Defaults to 1.
        confidence_threshold: The threshold for confidence-masking.
            It is a threshold of the probability in [0, 1],
            rather than of the logit. If set to -1, the threshold
            will be ignored. Defaults to -1.
        reduction: Default: 'mean'. This is the same as the `reduction`
            argument in
            :func:`texar.torch.losses.info_loss.kl_divg_loss_with_logits`.
            The loss will be a scalar tensor if the `reduction`
            is not :attr:`'none'`.
            Specifies the reduction to apply to the output:

            - ``'none'``: no reduction will be applied.
            - ``'batchmean'``: the sum of the output will be divided
              by the batch size.
            - ``'sum'``: the output will be summed.
            - ``'mean'``: the output will be divided by the number of elements
              in the output.
    """

    def __init__(
        self,
        sup_iterator: DataIterator,
        unsup_iterator: DataIterator,
        softmax_temperature: float = 1.0,
        confidence_threshold: float = -1,
        reduction: str = "mean",
    ):
        self._sup_iterator: DataIterator = sup_iterator
        self._unsup_iterator: DataIterator = unsup_iterator
        self._softmax_temperature = softmax_temperature
        self._confidence_threshold = confidence_threshold
        self._reduction = reduction

        # The flag for returning the unsupervised data.
        self._use_unsup = True
        self._sup_iter: Iterator[Batch]
        self._unsup_iter: Iterator[Batch]

    def __len__(self):
        return self._sup_iterator.__len__()

    def switch_to_dataset(
        self, dataset_name: Optional[str] = None, use_unsup: bool = True
    ):
        # Set the flag of using unsupervised data.
        self._use_unsup = use_unsup
        self._sup_iterator.switch_to_dataset(dataset_name)

    def switch_to_dataset_unsup(self, dataset_name: Optional[str] = None):
        self._unsup_iterator.switch_to_dataset(dataset_name)

    @property
    def num_datasets(self) -> int:
        return (
            self._sup_iterator.num_datasets + self._unsup_iterator.num_datasets
        )

    @property
    def dataset_names(self) -> List[str]:
        return (
            self._sup_iterator.dataset_names
            + self._unsup_iterator.dataset_names
        )

    def calculate_uda_loss(
        self, logits_orig: Tensor, logits_aug: Tensor
    ) -> Tensor:
        r"""
        This function calculate the KL divergence between the output
        probabilities of original input and augmented input. The two inputs
        should have the same shape, and the last dimension of them should be
        the probability distribution.

        Args:
            logits_orig: A tensor contains the logits of the original data.
            logits_aug: A tensor contains the logits of the augmented data.
                Must have the same shape as `logits_orig`.
        Returns:
            The loss, as a pytorch scalar float tensor if the `reduction`
            is not :attr:`'none'`, otherwise a tensor with the same shape
            as the `logits_orig`.
        """
        uda_loss = kl_divg_loss_with_logits(
            target_logits=logits_orig,
            input_logits=logits_aug,
            softmax_temperature=self._softmax_temperature,
            confidence_threshold=self._confidence_threshold,
            reduction=self._reduction,
        )
        return uda_loss

    def __iter__(self):
        r"""
        The :class:`texar.torch.data.DataIterator` is not
        inherited from the :class:`Iterator`. So we have to
        get the iterator explicitly.
        """
        self._sup_iter = iter(self._sup_iterator)
        self._unsup_iter = iter(self._unsup_iterator)
        return self

    def __next__(self) -> Tuple[Batch, Optional[Batch]]:
        r"""
        When :attr:`_use_unsup` = False, the iterator will be the same
        as a normal iterator for the supervised data. Otherwise,
        it will yield unsupervised batch, in addition to
        the supervised batch.

        The iterator will only raise the StopIteration when
        the supervised dataset reaches its end. The unsupervised
        data will always be yielded as long as the iterator
        still has supervised data.
        """
        if not self._use_unsup:
            return next(self._sup_iter), None
        try:
            sup_batch = next(self._sup_iter)
        except StopIteration as e:
            raise StopIteration from e
        try:
            unsup_batch = next(self._unsup_iter)
        except StopIteration:
            self._unsup_iter = iter(self._unsup_iterator)
            unsup_batch = next(self._unsup_iter)

        return sup_batch, unsup_batch
